# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

'''
RALEIGH (RAL EIGensolvers for real symmetric and Hermitian problems) core
solver.

For advanced users only - consider using more user-friendly interfaces in 
raleigh/interfaces first.

Implements a block Conjugate Gradient algorithm for the computation of 
several eigenpairs (eigenvalues and corresponding eigenvectors) of real 
symmetrtic or Hermitian problems, namely:
- standard eigenvalue problem 
  A x = lambda x
- generalized eigenvalue problems
  A x = lambda B x
  A B x = lambda x
where A and B are real symmetric or Hermitian operators, B being positive
definite.

The algorithm operates on sets of vectors (v_1, ..., v_m) encapsulated by
an abstract data type Vectors with the following methods:
---------------------------------
new_vectors(self, nv=0, dim=None)
  returns a new Vectors object encapsulating nv vectors of dimension dim
  if dim is not None or else the same dimension as self
---------------------------------
dimension(self):
  returns the dimension of vectors encapsulated by self
---------------------------------
select(self, nv, first=0)
  selects a subset of nv encapsulated vectors starting from first;
  all subsequent operations on self will involve these vectors only
---------------------------------
selected(self)
  returns current subset selection as tuple (first, nv)
---------------------------------
clone(self)
  returns a copy of (selected part of) self
---------------------------------
append(self, other):
  appends vectors from Vectors object other to those of self
---------------------------------
nvec(self)
  returns the number of currently selected vectors
---------------------------------
data_type(self)
  returns the data type of vectors' elements 
  (numpy.float32, numpy.float64, numpy.complex64 or numpy.complex128)
---------------------------------
fill_random(self):
  fills vectors with random values uniformly distributed between -1 and 1
---------------------------------
copy(self, other, ind=None)
  copies vectors from self to other:
  if ind is None, previously selected vectors of self copied into previously
  selected vectors of other (the numbers of selected vectors must coincide),
  otherwise vectors specified by the array of indices ind are copied
---------------------------------
scale(self, s, multiply=False)
  if multiply is True, previously selected vectors of self are multiplied by
  the respective elements of numpy ndarray s, otherwise the former are divided
  by the latter, skipping division for zero elements
---------------------------------
dots(self, other, transp=False)
  if transp is False, returns the numpy ndarray of dot products of previously
  selected vectors of self with the respective selected vectors of other,
  otherwise returns a numpy ndarray, i-th element of which is the dot product
  of the vector of i-th components of selected vectors of self by the vector
  of i-th components of selected vectors of other (note that for complex 
  vectors the complex dot products are computed, i.e. complex conjugation is
  applied to the components of vectors in other)
---------------------------------
dot(self, other)
  returns the ndarray of shape (m, n), where m and n are the numbers of the
  selected vectors in other and self respectively, containing dot products of
  the selected vectors of self by those of other (again, dot products are
  complex in the complex vectors case)
---------------------------------
multiply(self, q, other)
  for each column q[:, j] of ndarray q, assigns the linear combination of the
  selected vectors of self with coefficients q[0, j], q[1, j], ... to j-th
  selected vector of other
---------------------------------
add(self, other, s, q=None)
  if s is a scalar: 
    if q is None, adds the selected vectors of other multiplied by s to the 
    respective selected vectors of self, otherwise
    for each column q[:, j] of ndarray q, adds the linear combination of 
    selected vectors of other with coefficients q[0, j], q[1, j], ... 
    multiplied by s to j-th selected vector of self,
  otherwise adds the selected vectors of other multiplied by the respective
  elements of one-dimensional ndarray s to respective selected vectors of self
  (q is ignored)
---------------------------------

The folder raleigh/algebra contains three implementations of Vectors type:
(i) numpy implementaion, (ii) mkl implementation (requires MKL 10.3 or later:
needs mkl_rt.dll on Windows, libmkl_rt.so on Linux), and (iii) CUDA GPU 
implementation (requires CUDA-enabled GPU and NVIDIA Toolkit). These may be
used as templates for further implementations - MPI, out of core etc.

The number of wanted eigenpairs does not need to be set before calling the 
core solver - instead, the user may provide an object responsible for stopping 
the computation based on the data computed so far. 

An object responsible for deciding whether a particular eigenpair has converged 
can also be supplied by the user (default convergence criteria object is 
available).

If some eigenvectors are already available, they can be passed to the core 
solver, which then will compute further eigenpairs. Initial guesses to
eigenvectors may also be supplied by the user.
'''

import math
import numpy
import numpy.linalg as nla
import scipy.linalg as sla

RECORDS = 100


class DefaultConvergenceCriteria:
    '''
    Convergence criteria to be used if not specified by the user via Options
    (see Options.convergence_criteria below).
    '''
    def __init__(self):
        self.tolerance = 1e-3
        self.error = 'kinematic eigenvector error'
    def set_error_tolerance(self, error, tolerance):
        self.error = error
        self.tolerance = tolerance
    def satisfied(self, solver, i):
        err = solver.convergence_data(self.error, i)
        return err >= 0 and err <= self.tolerance


class Options:
    '''
    Solver options.
    
    Attributes
    ----------
    verbosity : int
        printout level
       <0 : no output
        0 : error an warning messages
        1 : iteration number, converged eigenvalues
        2 : convergence data for all current iterates
    max_iter : int
        maximal number of iterations per eigenpair;
        if negative, set by solver
    min_iter : int
        minimal number of iterations per eigenpair
    block_size : int
        the nuber of simultaneously iterated eigenvector approximations;
        if negative, set by solver
    threads : int
        the number of CPU threads, to be used to determine the block size if
        not set by the user
    sigma : float
        if not None, indicates that the solver is used in shift-invert context
    convergence_criteria : object
        if not None, must be an object with method satisfied(self, solver, i)
        that returns True if i-th approximate eigenpair converged and False
        otherwise, based on the information provided by solver.convergence_data
        (see below)
    stopping_criteria : object
        if not None, must be an object with method satisfied(self, solver)
        that returns True if sufficient number of eigenpairs have been computed
        and False otherwise, based on the values of solver attributes (e.g.
        solver.eigenvalues) and possibly the user input, see 
        interfaces.partial_svd.DefaultStoppingCriteria for an example
    detect_stagnation : bool
        if set to True, detects the loss of convergence, i.e. impossibility to
        significantly improve the accuracy of the approximation (set to False
        and request very high accuracy if you want to test the solver for 
        numerical stability)
    max_quota : float
        the iterations will stop as soon as the number of computed eigenpairs 
        exceeds max_quota multiplied by the problem size, and the rest of 
        eigenpairs will be computed by scipy.linalg.eigh
    '''
    def __init__(self):
        self.verbosity = 0
        self.max_iter = -1
        self.min_iter = 0
        self.block_size = -1
        self.threads = -1
        self.sigma = None
        self.convergence_criteria = None
        self.stopping_criteria = None
        self.detect_stagnation = True
        self.max_quota = 0.75


class EstimatedErrors:
    '''
    Estimated errors container.
    
    Attributes
    ----------
    kinematic : one-dimensional ndarray of floats
        error estimates based on the convergence history
    residual : one-dimensional ndarray of floats
        error estimates based on the residuals
    '''
    def __init__(self):
        self.kinematic = numpy.ndarray((0,), dtype=numpy.float32)
        self.residual = numpy.ndarray((0,), dtype=numpy.float32)
    def __getitem__(self, item):
        return self.kinematic[item], self.residual[item]
    def append(self, est):
        self.kinematic = numpy.concatenate((self.kinematic, est[0, :]))
        self.residual = numpy.concatenate((self.residual, est[1, :]))
    def reorder(self, ind):
        self.kinematic = self.kinematic[ind]
        self.residual = self.residual[ind]


class Problem:
    '''
    Eigenvalue problem specification.
    
    Attributes
    ----------
    __A : object
        operator A (stiffness matrix)
    __B : object
        operator B (mass matrix)
    __type : string
        problem type
        'std' :  standard A x = lambda x
        'gen' :  generalized A x = lambda B x
        'pro' :  generalized A B x = lambda x
    '''
    def __init__(self, v, A, B=None, prod=None):
        self.__vector = v
        self.__A = A
        self.__B = B
        if B is None:
            self.__type = 'std'
        else:
            if prod is None:
                self.__type = 'gen'
            else:
                self.__type = 'pro'
    def A(self):
        return self.__A
    def B(self):
        return self.__B
    def type(self):
        return self.__type[0]
    def vector(self):
        return self.__vector


class Solver:
    '''
    Eigenvalue problem solver specification.
    
    Attributes
    ----------
    __problem : instance of Problem
        problem specification
    __P : object
        preconditioner
    iteration : int
        the current iteration number
    lcon : int
        the number of computed eigenpairs on the left margin of the spectrum
    rcon : int
        the number of computed eigenpairs on the right margin of the spectrum
    eigenvalues : one-dimensional ndarray of dtype numpy.float64
        converged eigenvalues
    eigenvalue_errors : one-dimensional ndarray of dtype EstimatedErrors
        estimated errors for computed eigenvalues
    eigenvector_errors : one-dimensional ndarray of dtype EstimatedErrors
        estimated errors for computed eigenvectors
    residual_norms : one-dimensional ndarray of dtype numpy.float32
        residual norms for computed eigenpairs
    convergence_status : one-dimensional ndarray of dtype numpy.int32
        computed eigenpairs convergence status
        > 0 : the number of iterations taken to converge
        < 0 : the negative of the number of iterations taken to stagnate
    cnv : one-dimensional ndarray of dtype numpy.int32
        current convergence status
            0 : has not converged yet
        i > 0 : converged after i iterations
        i < 0 : stopped converging after i iterations
    lmd : one-dimensional ndarray of dtype numpy.float64
        current eigenvalue iterates
    res : one-dimensional ndarray of dtype numpy.float32
        current residuals
    err_lmd : two-dimensional ndarray of dtype numpy.float32
        error estimates for current eigenvalue iterates
    err_X : two-dimensional ndarray of dtype numpy.float32
        error estimates for current eigenvector iterates
    '''
    def __init__(self, problem):
        self.__problem = problem
        self.__P = None
        self.iteration = 0
        self.lcon = 0
        self.rcon = 0
        self.eigenvalues = numpy.ndarray((0,), dtype=numpy.float64)
        self.eigenvalue_errors = EstimatedErrors()
        self.eigenvector_errors = EstimatedErrors()
        self.residual_norms = numpy.ndarray((0,), dtype=numpy.float32)
        self.convergence_status = numpy.ndarray((0,), dtype=numpy.int32)
        # data to be set by solver
        self.eigenvectors = None
        self.eigenvectors_im = None
        self.block_size = None
        self.cnv = None
        self.lmd = None
        self.res = None
        self.err_lmd = None
        self.err_X = None

    def set_preconditioner(self, P):
        self.__P = P

    def problem(self):
        return self.__problem

    def preconditioner(self):
        return self.__P

    def convergence_data(self, what='residual', which=0):
        '''Reports current convergence data.
        
        Parameters
        ----------
        what : string
            convergence data to report (can be abbreviated, full names below)
        which : int
            for which eigenpair iterate to report
        '''
        if what.find('block') > -1:
            '''block size
            '''
            return self.block_size
        elif what.find('res') > -1 and what.find('vec') == -1:
            '''relative residual.
            
            WARNING: use only for the largest eigenvalues computation.
            '''
            max_lmd = numpy.amax(abs(self.lmd))
            if self.lcon + self.rcon > 0:
                max_lmd = max(max_lmd, numpy.amax(abs(self.eigenvalues)))
            return self.res[which]/max_lmd
        elif what.find('val') > -1:
            if what.find('max') > -1:
                max_lmd = numpy.amax(abs(self.lmd))
                if self.lcon + self.rcon > 0:
                    max_lmd = max(max_lmd, numpy.amax(abs(self.eigenvalues)))
                return max_lmd
            if what.find('err') > -1:
                err = self.err_lmd[:, which]
                if what.find('k'):
                    '''kinematic eigenvalue error estimate.
                    '''
                    return err[0]
                else:
                    '''residual-based eigenvalue error estimate.
                    '''
                    return err[1]
            else:
                '''current eigenvalue iterate.
                '''
                return self.lmd[which]
        elif what.find('vec') > -1:
            err = self.err_X[:, which]
            if what.find('k') > -1:
                '''kinematic eigenvector error estimate.
                '''
                return err[0]
            else:
                '''residual-based eigenvector error estimate.
                '''
                return err[1]
        else:
            raise ValueError('convergence data %s not found' % what)

    def solve(self, eigenvectors, options=Options(), which=(-1, -1), \
              extra=(-1, -1), init=(None, None)):
        '''Main core solver routine.
        
        Parameters
        ----------
        eigenvectors : object of abstract Vectors type
            eigenvectors container; normally empty (eigenvectors.nvec() = 0),
            if not, then assumed by the solver to contain previously computed
            eigenvectors of the solved problem; on return contains all computed
            eigenvectors (previous and new)
        options : object of type Options
            solver options
        which : int or tuple of two ints
            if int, the number of the largest eigenvalues wanted;
            if tuple, the numbers of eigenvalues wanted on the left (which[0])
            and right (which[1]) margine of the spectrum;
            negative values mark unknown number of wanted eigenvalues: in this
            case, the user must provide stopping_criteria
        extra : tuple of two ints
            numbers of extra eigenvectors corresponding to eigenvalues on the
            margins of the spectrum to iterate purely for the sake of better
            convergence: convergence criteria will not be applied to these
            extra eigenpairs, i.e. iterations will stop when all wanted
            eigenpairs converge
        init : tuple of two Vectors objects
            each tuple item, if not None, contains initial guesses for
            eigenvectors corresponding to eigenvalues on the respective margin
            of the spectrum
            
        Returns
        -------
        status : int
            execution status
            0 : success
            1 : maximal number of iterations exceeded
            2 : no search directions left (bad problem data or preconditioner)
            3 : some of the requested left eigenvalues may not exist
            4 : some of the requested right eigenvalues may not exist
           <0 : fatal error - exception thrown
        '''

        verb = options.verbosity

        try:
            l = len(which)
            if l != 2:
                raise ValueError\
                      ('which must be either integer or tuple of 2 integers')
            largest = False
        except:
            largest = True

        if largest:
            if which >= 0:
                left = which//2
                right = which - left
            else:
                left = -1
                right = -1
        else:
            left = int(which[0])
            right = int(which[1])
        if left == 0 and right == 0:
            if verb > -1:
                print('No eigenpairs requested, quit')
            return 0

        m = int(options.block_size)
        if m < 0:
            m = _default_block_size(left, right, extra, init, options.threads)
        else:
            if (left == 0 or right == 0) and not largest:
                if m < 3:
                    if verb > -1:
                        print('Block size %d too small, will use 3 instead' % m)
                    m = 3
            else:
                if m < 4:
                    if verb > -1:
                        print('Block size %d too small, will use 4 instead' % m)
                    m = 4
        self.block_size = m

        n = eigenvectors.dimension()

        #output
        self.iteration = 0
        self.lcon = 0
        self.rcon = 0
        self.eigenvalues = numpy.ndarray((0,), dtype=numpy.float64)
        self.eigenvalue_errors = EstimatedErrors()
        self.eigenvector_errors = EstimatedErrors()
        self.residual_norms = numpy.ndarray((0,), dtype=numpy.float32)
        self.convergence_status = numpy.ndarray((0,), dtype=numpy.int32)

        if m < n//2:
            try:
                status = self._solve(eigenvectors, options, which, extra, init)
                if status > 1:
                    if verb > -1:
                        print('core solver return status %d' % status)
                    return status - 1
            except _Error as err:
                if verb > -1:
                    print('%s' % err.value)
                return -1
        else:
            status = 1

        if status == 0:
            return 0 # success

        Xc = eigenvectors
        nc = Xc.nvec()
        m = n - nc
        if verb > -1:
            msg = '%d eigenpairs not computed by CG, applying ' + \
                  'Rayleigh-Ritz procedure'
            print( msg % m)
            print('in the complement subspace...')

        X = eigenvectors.new_vectors(m)
        X.fill_random()
        Y = X.new_vectors(m)
        Z = X.new_vectors(m)
        std = (self.__problem.type() == 's')
        pro = (self.__problem.type() == 'p')

        opA = self.problem().A()
        A = lambda x, y: opA.apply(x, y)
        opB = self.problem().B()
        if opB is not None:
            B = lambda x, y: opB.apply(x, y)
        data_type = eigenvectors.data_type()

        if nc > 0:
            if not std:
                BXc = eigenvectors.clone()
                if nc > 0:
                    B(Xc, BXc)
            else:
                BXc = Xc
            if nc > 0:
                Gc = BXc.dot(Xc)
                #Gci = nla.inv(Gc)
                # approximate inverse of Gc
                Gci = 2*numpy.identity(nc, dtype=data_type) - Gc
            Q = numpy.dot(Gci, X.dot(BXc))
            X.add(Xc, -1.0, Q)
            Q = numpy.dot(Gci, X.dot(BXc))
            X.add(Xc, -1.0, Q)

        if not std:
            B(X, Y)
            XBX = Y.dot(X)
        else:
            XBX = X.dot(X)
        lmd, q = sla.eigh(XBX)
        epsilon = 100*numpy.finfo(data_type).eps
        k = numpy.sum(lmd <= epsilon*lmd[-1])
        if k > 0:
            if verb > -1:
                #print(lmd[: k + 2], lmd[-1])
                msg = 'dropping %d linear dependent vectors ' + \
                      'from the Rayleigh-Ritz procedure...'
                print(msg % k)
            X.multiply(Q, Z)
            Z.copy(X)
            Y.multiply(Q, Z)
            Z.copy(Y)
            m -= k
            X.select(m)
            Y.select(m)
            Z.select(m)
            XBX = XBX[:m, :m]

        if pro:
            A(Y, Z)
            XAX = Z.dot(Y)
        else:
            A(X, Z)
            XAX = Z.dot(X)

        lmdx, Q = sla.eigh(XAX, XBX, turbo=False, overwrite_a=True, \
                           overwrite_b=True)
        X.multiply(Q, Z)
        Z.copy(X)
        eigenvectors.append(X)
        self.eigenvalues = numpy.concatenate((self.eigenvalues, lmdx))

        return 0

    def _solve(self, eigenvectors, options, which, extra, init):

        verb = options.verbosity
        sigma = options.sigma

        try:
            l = len(which)
            if l != 2:
                raise ValueError\
                      ('which must be either integer or tuple of 2 integers')
            largest = False
            left = int(which[0])
            right = int(which[1])
        except:
            largest = True
            left = which
            right = which

        m = self.block_size
        if left == 0 and not largest:
            r = 0.0
            l = 1
        elif right == 0:
            r = 1.0
            l = m - 1
        elif left > 0 and right > 0:
            r = left/(left + 1.0*right)
            l = int(round(r*m))
            if l < 2:
                l = 2
            if l > m - 2:
                l = m - 2
        else:
            r = 0.5
            l = m//2
        left_ratio = r
        block_size = m
        left_block_size = l

        extra_left = int(extra[0])
        extra_right = int(extra[1])
        if left >= 0:
            if extra_left > 0:
                left_total = left + extra_left
            else:
                left_total = max(left + 1, left_block_size)
            if verb > 2:
                print('left total: %d' % left_total)
        if right >= 0:
            if extra_right > 0:
                right_total = right + extra_right
            else:
                right_total = max(right + 1, block_size - left_block_size)
            if verb > 2:
                print('right_total: %d' % right_total)
        if verb > 0:
            print('left block size %d, right block size %d' % (l, m - l))

        # problem
        problem = self.__problem
        vector = problem.vector()
        problem_type = problem.type()
        std = (problem_type == 's')
        gen = (problem_type == 'g')
        pro = (problem_type == 'p')
        data_type = vector.data_type()
        epsilon = numpy.finfo(data_type).eps
        single = (data_type == numpy.float32 or data_type == numpy.complex64)

        # convergence data
        self.cnv = numpy.zeros((m,), dtype=numpy.int32)
        self.lmd = numpy.zeros((m,), dtype=numpy.float64)
        self.res = -numpy.ones((m,), dtype=numpy.float32)
        self.err_lmd = -numpy.ones((2, m,), dtype=numpy.float32)
        self.err_X = -numpy.ones((2, m,), dtype=numpy.float32)

        # convergence criteria
        if options.convergence_criteria is None:
            convergence_criteria = DefaultConvergenceCriteria()
        else:
            convergence_criteria = options.convergence_criteria

        # convergence history data
        iterations = numpy.zeros((m,), dtype=numpy.int32)
        dlmd = numpy.zeros((m, RECORDS), dtype=numpy.float32)
        dX = numpy.ones((m,), dtype=numpy.float32)
        acf = numpy.ones((2, m,), dtype=numpy.float32)
        cluster = numpy.zeros((2, m), dtype=numpy.int32)

        # workspace
        X = vector.new_vectors(m)
        X.fill_random()
        Y = vector.new_vectors(m)
        Z = vector.new_vectors(m)
        W = vector.new_vectors(m)
        AX = vector.new_vectors(m)
        AY = vector.new_vectors(m)
        if not std:
            BX = vector.new_vectors(m)
            BY = vector.new_vectors(m)
        else:
            BX = X
            BY = Y
        AZ = AY
        BZ = BY

        # copy initial vectors if present
        l = left_block_size
        m = block_size
        init_lX = init[0]
        if init_lX is not None:
            init_left = min(l, init_lX.nvec())
            X.select(init_left)
            init_lX.select(init_left)
            init_lX.copy(X)
        else:
            init_left = 0
        init_rX = init[1]
        if init_rX is not None:
            init_right = min(m - l, init_rX.nvec())
            X.select(init_right, init_left)
            init_rX.select(init_right)
            init_rX.copy(X)

        # check for zero initial vectors
        X.select(m)
        s = X.dots(X)
        for i in range(m):
            if s[i] == 0.0:
                if verb > -1:
                    print('Zero initial guess, replacing with random')
                X.select(1, i)
                X.fill_random()
                s[i : i + 1] = X.dots(X)
        X.select(m)
        s = numpy.sqrt(X.dots(X))
        X.scale(s)

        # shorcuts
        detect_stagn = options.detect_stagnation
        lmd = self.lmd
        res = self.res
        err_lmd = self.err_lmd
        err_X = self.err_X
        A = lambda x, y: problem.A().apply(x, y)
        opB = problem.B()
        if opB is not None:
            B = lambda x, y: opB.apply(x, y)
        else:
            B = None
        opP = self.__P
        if opP is not None:
            P = lambda x, y: opP.apply(x, y)
        else:
            P = None

        # constraints (already available eigenvectors e.g. from previous run)
        self.eigenvectors = eigenvectors
        Xc = eigenvectors
        nc = Xc.nvec()
        if not std:
            BXc = eigenvectors.clone()
            if nc > 0:
                B(Xc, BXc)
            self.eigenvectors_im = BXc
        else:
            BXc = Xc
        if nc > 0:
            Gc = BXc.dot(Xc)
            # approximate inverse of Gc
            Gci = 2*numpy.identity(nc, dtype=data_type) - Gc

        # initialize
        leftX = left_block_size
        rightX = block_size - leftX
        rec = 0
        ix = 0 # first X
        nx = block_size
        ny = block_size
        nz = 0
        lmdz = None

        if Xc.nvec() > 0:
            # orthogonalize X to Xc
            # std: X := X - Xc Xc* X
            # gen: X := X - Xc BXc* X
            # pro: X := X - Xc BXc* X
            Q = numpy.dot(Gci, X.dot(BXc))
            X.add(Xc, -1.0, Q)

        if not std:
            B(X, BX)
        XBX = BX.dot(X)

        # do pivoted Cholesky for XBX to eliminate linear dependent X
        U = XBX.copy()
        ind, dropped = _piv_chol(U, 0, 1e-2)
        if dropped > 0:
            if verb > 0:
                print('dropped %d initial vectors out of %d' % (dropped, nx))
            # drop linear dependent initial vectors
            nx -= dropped
            if nx > 0:
                W.select(nx)
                X.copy(W, ind)
                W.copy(X)
            X.select(dropped, nx)
            X.fill_random()
            if not std:
                if nx > 0:
                    BX.copy(W, ind)
                    W.copy(BX)
                BX.select(dropped, nx)
                B(X, BX)
            if Xc.nvec() > 0:
                # orthogonalize X to Xc
                Q = numpy.dot(Gci, X.dot(BXc))
                Xc.multiply(Q, W)
                X.add(W, -1.0)
                if not std:
                    BXc.multiply(Q, W)
                    BX.add(W, -1.0)
            nx = m
            X.select(nx)
            if not std:
                BX.select(nx)
            XBX = BX.dot(X)

        # Rayleigh-Ritz in the initial space
        if pro:
            A(BX, AX)
            XAX = AX.dot(BX)
        else:
            A(X, AX)
            XAX = AX.dot(X)
        lmdx, Q = sla.eigh(XAX, XBX, turbo=False)
        W.select(m)
        X.multiply(Q, W)
        W.copy(X)
        AX.multiply(Q, W)
        W.copy(AX)
        if not std:
            BX.multiply(Q, Z)
            Z.copy(BX)

# ===== main CG loop
        max_iter = options.max_iter
        min_iter = options.min_iter
        if max_iter < 0:
            max_iter = 100
        self.iteration = 0
        while True:

            maxit = 0
            if left != 0 and left_block_size > 0:
                maxit = numpy.amax(iterations[:left_block_size])
            if right != 0 and left_block_size < block_size:
                maxit = max(maxit, numpy.amax(iterations[left_block_size:]))
            if maxit >= max_iter:
                if verb > -1:
                    msg = 'iterations limit of %d exceeded, terminating'
                    print(msg % max_iter)
                break

            if verb > 0:
                print('------------- iteration %d' % self.iteration)

            if pro:
                XAX = AX.dot(BX)
            else:
                XAX = AX.dot(X)
            XBX = BX.dot(X)
            da = XAX.diagonal()
            db = XBX.diagonal()
            new_lmd = _real(da/db)

            # estimate error in residual computation due to the error in
            # computing AX, to be used in detecting convergence stagnation
            Lmd = numpy.zeros((nx, nx))
            Lmd[range(nx), range(nx)] = lmdx #new_lmd
            RX = XAX - numpy.dot(XBX, Lmd)
            delta_R = _norm(RX, 0)
            if gen:
                s = numpy.sqrt(abs(X.dots(X)))
                delta_R /= s
            rv_err = numpy.amax(abs(new_lmd - lmdx))
            if verb > 2:
                print('Ritz values error: %.1e' % rv_err)
                print('Ritz vectors non-orthonormality: %.1e' % \
                      numpy.amax(abs(XBX - numpy.eye(nx))))

            if self.iteration > 0:
                # compute eigenvalue decrements
                for i in range(nx):
                    if iterations[ix + i]:
                        delta = lmd[ix + i] - new_lmd[i]
                        eps = math.sqrt(epsilon)
                        eps *= max(abs(lmd[ix + i]), abs(new_lmd[i]))
                        if abs(delta) > eps:
                            dlmd[ix + i, rec - 1] = delta
                    iterations[ix + i] += 1
                if verb > 3:
                    print('eigenvalues shifts history:')
                    print(numpy.array_str(dlmd[ix : ix + nx, :rec].T, \
                                          precision=2))

            lmd[ix : ix + nx] = new_lmd

            # compute residuals
            # std: A X - X lmd
            # gen: A X - B X lmd
            # pro: A B X - X lmd
            W.select(nx, ix)
            Y.select(nx)
            AX.copy(W)
            if gen:
                W.add(BX, -lmd[ix : ix + nx])
            else:
                W.add(X, -lmd[ix : ix + nx])

            if Xc.nvec() > 0:
                # orthogonalize W to Xc
                # std: W := W - Xc Xc* W
                # gen: W := W - BXc Xc* W
                # pro: W := W - Xc BXc* W
                if pro:
                    Q = numpy.dot(Gci, W.dot(BXc))
                else:
                    Q = numpy.dot(Gci, W.dot(Xc))
                if gen:
                    W.add(BXc, -1.0, Q)
                else:
                    W.add(Xc, -1.0, Q)

            if pro:
                W.copy(Y)
                B(Y, W)
                s = W.dots(Y)
            else:
                s = W.dots(W)
            res[ix : ix + nx] = numpy.sqrt(abs(s))

            # kinematic error estimates
            if rec > 3: # sufficient history available
                for i in range(nx):
                    if dX[ix + i] > 0.01:
                        err_X[0, ix + i] = -1.0
                        continue
                    k = 0
                    s = 0
                    # go through the last 1/3 of the history
                    for r in range(rec - 1, rec - rec//3 - 2, -1):
                        d = abs(dlmd[ix + i, r])
                        if d == 0:
                            break
                        k = k + 1
                        s = s + d
                    if k < 2 or s == 0:
                        continue
                    # estimate asymptotic convergence factor (a.c.f)
                    qi = abs(dlmd[ix + i, rec - 1])/s
                    if qi <= 0:
                        continue
                    qi = qi**(1.0/(k - 1))
                    acf[1, ix + i] = acf[0, ix + i]
                    acf[0, ix + i] = qi # a.c.f. estimate
                    if qi >= 1.0:
                        continue
                    # esimate error based on a.c.f.
                    theta = qi/(1 - qi)
                    d = theta*dlmd[ix + i, rec - 1]
                    err_lmd[0, ix + i] = abs(d)
                    qx = math.sqrt(qi)
                    err_X[0, ix + i] = dX[ix + i]*qx/(1 - qx)

            if not gen:
                # residual-based error estimates:
                # asymptotic Lehmann for eigenvalues
                # generalized (extended gap) Davis-Kahan for eigenvectors;
                # not valid for the generalized eigenvalue problem
                l = 0
                for k in range(1, leftX):
                    i = ix + k
                    if dX[i] > 0.01:
                        break
                    if lmd[i] - lmd[i - 1] > res[i]:
                        l = k
                if l > 0:
                    i = ix + l
                    t = lmd[i]
                    if verb > 2:
                        print('using left pole at lmd[%d] = %e' % (i, t))
                    m = block_size
                    for k in range(l):
                        i = ix + k
                        s = res[i]
                        err_lmd[1, i] = s*s/(t - lmd[i])
                        err_X[1, i] = s/(t - lmd[i])
                l = 0
                for k in range(1, rightX):
                    i = ix + nx - k - 1
                    if dX[i] > 0.01:
                        break
                    if lmd[i + 1] - lmd[i] > res[i]:
                        l = k
                if l > 0:
                    i = ix + nx - l - 1
                    t = lmd[i]
                    if verb > 2:
                        print('using right pole at lmd[%d] = %e' % (i, t))
                    m = block_size
                    for k in range(l):
                        i = ix + nx - k - 1
                        s = res[i]
                        err_lmd[1, i] = s*s/(lmd[i] - t)
                        err_X[1, i] = s/(lmd[i] - t)

            if verb > 1:
                msg = '  eigenvalue   residual   ' + \
                'estimated errors (kinematic/residual)' + \
                '      a.c.f.'
                print(msg)
                msg = '                          ' + \
                '   eigenvalue            eigenvector '
                print(msg)
                for i in range(block_size):
                    print('%14e %8.1e  %8.1e / %8.1e    %.1e / %.1e  %.3e  %d' % \
                          (lmd[i], res[i], \
                          err_lmd[0, i], err_lmd[1, i], \
                          abs(err_X[0, i]), abs(err_X[1, i]), \
                          acf[0, i], self.cnv[i]))

            if self.iteration >= 2:
                eps = epsilon**0.67
                lbs = left_block_size
                dlmd_min_lft = eps*numpy.amax(abs(dlmd[:lbs, rec - 1]))
                dlmd_min_rgt = eps*numpy.amax(abs(dlmd[lbs:, rec - 1]))
                if self.iteration == 2:
                    dlmd_min_left = dlmd_min_lft
                    dlmd_min_right =  dlmd_min_rgt

            if self.iteration >= 2:
                cluster[:, :] = 0
                nc = 0
                for i in range(left_block_size - 1):
                    if abs(lmd[i + 1] - lmd[i]) <= dlmd_min_lft:
                        if cluster[0, i] == 0:
                            nc += 1
                            cluster[0, i] = nc
                            cluster[1, i] = 1
                        cluster[0, i + 1] = cluster[0, i]
                        cluster[1, i + 1] = cluster[1, i] + 1
                for j in range(m - left_block_size - 1):
                    i = m - j - 1
                    if abs(lmd[i - 1] - lmd[i]) <= dlmd_min_rgt:
                        if cluster[0, i] == 0:
                            nc += 1
                            cluster[0, i] = nc
                            cluster[1, i] = 1
                        cluster[0, i - 1] = cluster[0, i]
                        cluster[1, i - 1] = cluster[1, i] + 1
                if verb > 2:
                    print(cluster[0, :])
                    print(cluster[1, :])

            lcon = 0
            for i in range(leftX - leftX//4):
                if left == 0:
                    break
                j = self.lcon + i
                k = ix + i
                if sigma is not None and lmd[k] > 0:
                    break
                it = iterations[k]
                if it < min_iter:
                    break
                dlmd1 = abs(dlmd[k, max(0, rec - 1)])
                dlmd2 = abs(dlmd[k, max(0, rec - 3)])
                if convergence_criteria.satisfied(self, k):
                    if verb > 0:
                        msg = 'left eigenpair %d converged' + \
                        ' after %d iterations,\n' + \
                        ' eigenvalue %e, error %.1e / %.1e'
                        it = iterations[k]
                        print(msg % (j, it, lmd[k], err_X[0, k], err_X[1, k]))
                    lcon += 1
                    self.cnv[k] = self.iteration + 1
                elif detect_stagn and it > 2 and dlmd1 <= dlmd_min_left \
                     and (dlmd1 > dlmd2 or dlmd1 == 0.0):
                    if verb > 0:
                        msg = 'left eigenpair %d stagnated,\n' + \
                        ' eigenvalue %e, error %.1e / %.1e'
                        print(msg % (j, lmd[k], err_X[0, k], err_X[1, k]))
                    lcon += 1
                    self.cnv[k] = -self.iteration - 1
                else:
                    if cluster[0, k] > 0:
                        for l in range(k - 1, k - cluster[1, k], -1):
                            if self.cnv[l] == -self.iteration - 1:
                                self.cnv[l] = 0
                                lcon -= 1
                                if verb > 0:
                                    msg = 'stagnation of %e cancelled'
                                    print(msg % lmd[l])
                    break

            rcon = 0
            for i in range(rightX - rightX//4):
                if right == 0:
                    break
                j = self.rcon + i
                k = ix + nx - i - 1
                if sigma is not None and lmd[k] < 0:
                    break
                it = iterations[k]
                if it < min_iter:
                    break
                dlmd1 = abs(dlmd[k, max(0, rec - 1)])
                dlmd2 = abs(dlmd[k, max(0, rec - 3)])
                if convergence_criteria.satisfied(self, k):
                    if verb > 0:
                        msg = 'right eigenpair %d converged' + \
                        ' after %d iterations,\n' + \
                        ' eigenvalue %e, residual %.1e, error %.1e / %.1e'
                        print(msg % (j, it, lmd[k], res[k], err_X[0, k], err_X[1, k]))
                    rcon += 1
                    self.cnv[k] = self.iteration + 1
                elif detect_stagn and it > 2 and dlmd1 <= dlmd_min_right \
                     and (dlmd1 > dlmd2 or dlmd1 == 0.0):
                    if verb > 0:
                        msg = 'right eigenpair %d stagnated,\n' + \
                        ' eigenvalue %e, error %.1e / %.1e'
                        print(msg % (j, lmd[k], err_X[0, k], err_X[1, k]))
                    rcon += 1
                    self.cnv[k] = -self.iteration - 1
                else:
                    if cluster[0, k] > 0:
                        for l in range(k + 1, k + cluster[1, k]):
                            if self.cnv[l] == -self.iteration - 1:
                                self.cnv[l] = 0
                                rcon -= 1
                                if verb > 0:
                                    msg = 'stagnation of %e cancelled'
                                    print(msg % lmd[l])
                    break

            if largest: # ensure the largest converge first
                if lcon > 0:
                    i = ix + lcon - 1
                    j = ix + nx - rcon - 1
                    while lcon > 0 and abs(lmd[i]) < abs(lmd[j]):
                        self.cnv[i] = 0
                        lcon -= 1
                        i -= 1
                if rcon > 0:
                    i = ix + lcon
                    j = ix + nx - rcon
                    while rcon > 0 and abs(lmd[i]) > abs(lmd[j]):
                        self.cnv[j] = 0
                        rcon -= 1
                        j += 1

            # move converged X to Xc, update Gram matrix for Xc
            ncon = Xc.nvec()
            if lcon > 0:
                self.eigenvalues = numpy.concatenate \
                    ((self.eigenvalues, lmd[ix : ix + lcon]))
                self.eigenvalue_errors.append(err_lmd[:, ix : ix + lcon])
                self.eigenvector_errors.append(err_X[:, ix : ix + lcon])
                self.residual_norms = numpy.concatenate \
                    ((self.residual_norms, res[ix : ix + lcon]))
                self.convergence_status = numpy.concatenate \
                    ((self.convergence_status, self.cnv[ix : ix + lcon]))
                X.select(lcon, ix)
                if std and ncon > 0:
                    if ncon > 0:
                        Gu = X.dot(Xc)
                Xc.append(X)
                if not std:
                    if ncon > 0:
                        Gu = X.dot(BXc)
                    BX.select(lcon, ix)
                    BXc.append(BX)
                    if ncon < 1:
                        Gc = BXc.dot(Xc)
                    else:
                        Gl = BXc.dot(X)
                else:
                    if ncon < 1:
                        Gc = Xc.dot(Xc)
                    else:
                        Gl = Xc.dot(X)
                if ncon > 0:
                    Gc = numpy.concatenate((Gc, Gu), axis=1)
                    Gc = numpy.concatenate((Gc, Gl))
                ncon += lcon
            if rcon > 0:
                jx = ix + nx
                self.eigenvalues = numpy.concatenate \
                    ((self.eigenvalues, lmd[jx - rcon : jx]))
                self.eigenvalue_errors.append(err_lmd[:, jx - rcon : jx])
                self.eigenvector_errors.append(err_X[:, jx - rcon : jx])
                self.residual_norms = numpy.concatenate \
                    ((self.residual_norms, res[jx - rcon : jx]))
                self.convergence_status = numpy.concatenate \
                    ((self.convergence_status, self.cnv[jx - rcon : jx]))
                X.select(rcon, jx - rcon)
                if std and ncon > 0:
                    if ncon > 0:
                        Gu = X.dot(Xc)
                Xc.append(X)
                if not std:
                    if ncon > 0:
                        Gu = X.dot(BXc)
                    BX.select(rcon, jx - rcon)
                    BXc.append(BX)
                    if ncon < 1:
                        Gc = BXc.dot(Xc)
                    else:
                        Gl = BXc.dot(X)
                else:
                    if ncon < 1:
                        Gc = Xc.dot(Xc)
                    else:
                        Gl = Xc.dot(X)
                if ncon > 0:
                    Gc = numpy.concatenate((Gc, Gu), axis=1)
                    Gc = numpy.concatenate((Gc, Gl))
                ncon += rcon
            if ncon > 0:
                H = Gc - numpy.identity(ncon, dtype=data_type)
                if verb > 2:
                    print('Gram error: %e' % nla.norm(H))
                # approximate inverse, good enough if Gram offdiagonal
                # entries are less than sqrt(epsilon)
                Gci = 2*numpy.identity(ncon, dtype=data_type) - Gc

            self.lcon += lcon
            self.rcon += rcon
            if options.stopping_criteria is not None:
                if options.stopping_criteria.satisfied(self):
                    return 0
            if largest and right > 0 and self.lcon + self.rcon >= right:
                return 0
            left_converged = left >= 0 and self.lcon >= left
            right_converged = right >= 0 and self.rcon >= right
            if left_converged and right_converged:
                return 0
            if sigma is not None:
                if right_converged:
                    i = ix + lcon
                    lmd_i = lmd[i]
                    err_i = err_lmd[0, i]
                    if lmd_i > 0 and err_i != -1.0 and err_i < lmd_i/4:
                        return 4
                if left_converged:
                    i = ix + nx - rcon - 1
                    lmd_i = lmd[i]
                    err_i = err_lmd[0, i]
                    if lmd_i < 0 and err_i != -1.0 and err_i < -lmd_i/4:
                        return 5
            lim = options.max_quota * eigenvectors.dimension()
            if eigenvectors.nvec() > lim:
                return 1

            leftX -= lcon
            rightX -= rcon

            # re-select Xs, AXs, BXs accordingly
            iy = ix
            ny = nx
            ix += lcon
            nx -= lcon + rcon
            X.select(nx, ix)
            AX.select(nx, ix)
            if not std:
                BX.select(nx, ix)
            XAX = XAX[lcon : lcon + nx, lcon : lcon + nx]
            XBX = XBX[lcon : lcon + nx, lcon : lcon + nx]

            if not pro:
                if P is None:
                    W.copy(Y)
                else:
                    P(W, Y)

            if nz > 0:
                # compute the conjugation matrix
                if pro:
                    ZAY = W.dot(AZ)
                else:
                    ZAY = Y.dot(AZ)
                if std:
                    ZBY = Y.dot(Z)
                else:
                    ZBY = Y.dot(BZ)
                Num = ZAY - numpy.dot(ZBY, numpy.diag(lmd[iy : iy + ny]))
                ny = Y.nvec()
                Lmd = numpy.ndarray((1, ny))
                Mu = numpy.ndarray((nz, 1))
                Lmd[0, :] = lmd[iy : iy + ny]
                Mu[:, 0] = lmdz
                Den = Mu - Lmd
                sy = numpy.sqrt(abs(Y.dots(Y)))
                sz = numpy.sqrt(abs(Z.dots(Z)))
                Beta = numpy.ndarray((nz, ny), dtype=data_type)
                for iz in range(nz):
                    for iy in range(ny):
                        s = sy[iy]/sz[iz]
                        if abs(Num[iz, iy]) >= 100*s*abs(Den[iz, iy]):
                            Beta[iz, iy] = 0.0
                        else:
                            Beta[iz, iy] = Num[iz, iy]/Den[iz, iy]

                # conjugate search directions
                AZ.select(ny)
                Y.add(Z, -1.0, Beta)
                if pro: # if gen or nz == 0, BY computed later
                    W.add(BZ, -1.0, Beta)
                    BY.select(ny)
                    W.copy(BY)
            elif pro:
                BY.select(ny)
                W.copy(BY)

            Q = Y.dot(BX)
            Y.add(X, -1.0, Q)
            if pro:
                BY.add(BX, -1.0, Q)

            if Xc.nvec() > 0:
                # orthogonalize Y to Xc
                # std: W := W - Xc Xc* W (not needed if P is None)
                # gen: W := W - Xc BXc* W
                # pro: W := W - Xc BXc* W (not needed if P is None)
                Q = numpy.dot(Gci, Y.dot(BXc))
                Y.add(Xc, -1.0, Q)
                if pro:
                    BY.add(BXc, -1.0, Q)

            # compute (B-)Gram matrix for (X,Y)
            if std:
                s = numpy.sqrt(abs(Y.dots(Y)))
                Y.scale(s)
                if nx > 0:
                    XBY = Y.dot(X)
                YBY = Y.dot(Y)
            else:
                BY.select(Y.nvec())
                if not pro: # or nz == 0:
                    B(Y, BY)
                s = numpy.sqrt(abs(BY.dots(Y)))
                Y.scale(s)
                BY.scale(s)
                if nx > 0:
                    XBY = BY.dot(X)
                YBY = BY.dot(Y)

            if nx > 0:
                YBX = _conjugate(XBY)
                GB = numpy.concatenate((XBX, YBX))
                H = numpy.concatenate((XBY, YBY))
                GB = numpy.concatenate((GB, H), axis=1)
            else:
                GB = YBY

            # do pivoted Cholesky for GB
            U = GB
            ny = Y.nvec()
            if single:
                eps = 1e-4
            else:
                eps = 1e-8
            ind, dropped = _piv_chol(U, nx, eps)
            if dropped > 0:
                if verb > 0:
                    print('dropped %d search directions out of %d' \
                          % (dropped, ny))

            ny -= dropped
            if ny < 1:
                if verb > -1:
                    print('no search directions left, terminating')
                return 3

            # re-arrange/drop-linear-dependent search directions
            nxy = nx + ny
            U = U[:nxy, :nxy]

            indy = ind[nx: nxy]
            for i in range(ny):
                indy[i] -= nx
            W.select(ny)
            Y.copy(W, indy[:ny])
            Y.select(ny)
            W.copy(Y)
            AY.select(ny)
            if not std:
                BY.copy(W, indy[:ny])
                BY.select(ny)
                W.copy(BY)

            # compute A-Gram matrix for (X,Y)
            if pro:
                A(BY, AY)
                if nx > 0:
                    XAY = AY.dot(BX)
                YAY = AY.dot(BY)
            else:
                A(Y, AY)
                if nx > 0:
                    XAY = AY.dot(X)
                YAY = AY.dot(Y)
            if nx > 0:
                YAX = _conjugate(XAY)
                GA = numpy.concatenate((XAX, YAX))
                H = numpy.concatenate((XAY, YAY))
                GA = numpy.concatenate((GA, H), axis=1)
            else:
                GA = YAY

            # solve Rayleigh-Ritz eigenproblem
            G = _transform(GA, U)
            YAY = G[nx : nxy, nx : nxy]
            lmdy, Qy = sla.eigh(YAY)
            G[:, nx : nxy] = numpy.dot(G[:, nx : nxy], Qy)
            if nx > 0:
                G[nx : nxy, :nx] = _conjugate(G[:nx, nx : nxy])
            G[nx : nxy, nx : nxy] = numpy.dot(_conjugate(Qy), G[nx : nxy, nx : nxy])

            if G.dtype.kind == 'c':
                G = G.astype(numpy.complex128)
            else:
                G = G.astype(numpy.float64)

            lmdxy, Q = sla.eigh(G, turbo = False)

            lmdxy = lmdxy.astype(lmdy.dtype)
            Q = Q.astype(Qy.dtype)

            # estimate changes in eigenvalues and eigenvectors
            lmdx = numpy.concatenate \
                ((lmdxy[:leftX], lmdxy[nxy - rightX:]))
            lmdy = lmdxy[leftX : nxy - rightX]
            QX = numpy.concatenate \
                ((Q[:, :leftX], Q[:, nxy - rightX:]), axis=1)
            QYX = QX[nx:, :].copy()
            lmdX = numpy.ndarray((1, nx))
            lmdY = numpy.ndarray((ny, 1))
            lmdX[0, :] = lmdx
            lmdY[:, 0] = lmdy
            Delta = (lmdY - lmdX)*QYX*QYX
            dX[ix : ix + nx] = _norm(QYX, 0)
            if rec == RECORDS:
                for i in range(rec - 1):
                    dlmd[:, i] = dlmd[:, i + 1]
            else:
                rec += 1
            dlmd[ix : ix + nx, rec - 1] = _real(numpy.sum(Delta, axis=0))

            # select new numbers of left and right eigenpairs
            if left < 0:
                shift_left = ix
            elif lcon > 0:
                shift_left = max(0, left_total - self.lcon - leftX)
                shift_left = min(shift_left, ix)
            else:
                shift_left = 0
            if right < 0:
                shift_right = block_size - ix - nx
            elif rcon > 0:
                shift_right = max(0, right_total - self.rcon - rightX)
                shift_right = min(shift_right, block_size - ix - nx)
            else:
                shift_right = 0
            if shift_left + shift_right > ny:
                shift_left = min(shift_left, int(round(left_ratio*ny)))
                shift_right = min(shift_right, ny - shift_left)
            if left > 0 and lcon > 0 and self.lcon >= left:
                if verb > 0:
                    print('left-hand side converged')
                leftX_new = 0
                l = left_block_size
                rightX_new = min(nxy, l + rightX + shift_right)
                left_block_size_new = l + rightX + shift_right - rightX_new
                shift_left = -leftX - lcon
                left_ratio = 0.0
                ix_new = left_block_size_new
            elif right > 0 and rcon > 0 and self.rcon >= right:
                if verb > 0:
                    print('right-hand side converged')
                ix_new = ix - shift_left
                leftX_new = min(nxy, block_size - ix_new)
                rightX_new = 0
                shift_right = -rightX - rcon
                left_block_size_new = ix_new + leftX_new
                left_ratio = 1.0
            else:
                leftX_new = leftX + shift_left
                rightX_new = rightX + shift_right
                left_block_size_new = left_block_size
                ix_new = ix - shift_left
            nx_new = leftX_new + rightX_new
            if verb > 2:
                print('left X: was %d, now %d' % (leftX, leftX_new))
                print('right X: was %d, now %d' % (rightX, rightX_new))
                print('new ix %d, new nx %d, nxy %d' % (ix_new, nx_new, nxy))

            # shift eigenvalues etc.
            m = block_size
            l = left_block_size
            nl = left_block_size_new
            cnv = self.cnv
            if shift_left > 0:
                for i in range(l - shift_left):
                    j = i + shift_left
                    cnv[i] = cnv[j]
                    lmd[i] = lmd[j]
                    res[i] = res[j]
                    acf[:, i] = acf[:, j]
                    err_lmd[:, i] = err_lmd[:, j]
                    dlmd[i, :] = dlmd[j, :]
                    err_X[:, i] = err_X[:, j]
                    dX[i] = dX[j]
                    iterations[i] = iterations[j]
            if shift_left >= 0:
                for i in range(l - shift_left, nl):
                    _reset_cnv_data \
                        (i, cnv, res, acf, err_lmd, dlmd, err_X, dX, iterations)
            else:
                for i in range(l):
                    _reset_cnv_data \
                        (i, cnv, res, acf, err_lmd, dlmd, err_X, dX, iterations)
            if shift_right > 0:
                for i in range(m - 1, l + shift_right - 1, -1):
                    j = i - shift_right
                    cnv[i] = cnv[j]
                    lmd[i] = lmd[j]
                    res[i] = res[j]
                    acf[:, i] = acf[:, j]
                    err_lmd[:, i] = err_lmd[:, j]
                    dlmd[i, :] = dlmd[j, :]
                    err_X[:, i] = err_X[:, j]
                    dX[i] = dX[j]
                    iterations[i] = iterations[j]
            if shift_right >= 0:
                for i in range(l + shift_right - 1, nl - 1, -1):
                    _reset_cnv_data \
                        (i, cnv, res, acf, err_lmd, dlmd, err_X, dX, iterations)
            else:
                for i in range(l, block_size):
                    _reset_cnv_data \
                        (i, cnv, res, acf, err_lmd, dlmd, err_X, dX, iterations)

            # compute RR coefficients for X and 'old search directions' Z
            # by re-arranging columns of Q
            Q[nx : nxy, :] = numpy.dot(Qy, Q[nx : nxy, :])
            Q = sla.solve_triangular(U, Q)
            lmdx = numpy.concatenate \
                ((lmdxy[:leftX_new], lmdxy[nxy - rightX_new:]))
            QX = numpy.concatenate \
                ((Q[:, :leftX_new], Q[:, nxy - rightX_new:]), axis=1)
            lft = leftX_new
            rgt = rightX_new
            nz = nxy - lft - rgt
            lmdz = lmdxy[lft : nxy - rgt]
            QZ = Q[:, lft : nxy - rgt]
            if nx > 0:
                QXX = QX[:nx, :].copy()
            QYX = QX[nx:, :].copy()
            if nx > 0:
                QXZ = QZ[:nx, :].copy()
            QYZ = QZ[nx:, :].copy()

            # update X and 'old search directions' Z and their A- and B-images
            W.select(nx_new)
            Z.select(nx_new)
            if nx > 0:
                AX.multiply(QXX, W)
                W.add(AY, 1.0, QYX)
            else:
                AY.multiply(QYX, W)
            if nz > 0:
                Z.select(nz)
                AY.multiply(QYZ, Z)
                AZ.select(nz)
                if nx > 0:
                    Z.add(AX, 1.0, QXZ)
                Z.copy(AZ)
            AX.select(nx_new, ix_new)
            W.copy(AX)
            if not std:
                Z.select(nx_new)
                if nx > 0:
                    BX.multiply(QXX, W)
                    W.add(BY, 1.0, QYX)
                else:
                    BY.multiply(QYX, W)
                if nz > 0:
                    Z.select(nz)
                    BY.multiply(QYZ, Z)
                    BZ.select(nz)
                    if nx > 0:
                        Z.add(BX, 1.0, QXZ)
                    Z.copy(BZ)
                BX.select(nx_new, ix_new)
                W.copy(BX)
            else:
                BZ = Z
            Z.select(nx_new)
            if nx > 0:
                X.multiply(QXX, W)
                W.add(Y, 1.0, QYX)
            else:
                Y.multiply(QYX, W)
            if nz > 0:
                Z.select(nz)
                Y.multiply(QYZ, Z)
                if nx > 0:
                    Z.add(X, 1.0, QXZ)
            X.select(nx_new, ix_new)
            W.copy(X)

            nx = nx_new
            ix = ix_new
            leftX = leftX_new
            rightX = rightX_new
            left_block_size = left_block_size_new
            self.iteration += 1

        return 2

class _Error(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return '??? ' + repr(self.value)

def _conjugate(a):
    if a.dtype.kind == 'c':
        return a.conj().T
    else:
        return a.T

def _real(a):
    if a.dtype.kind == 'c':
        return a.real
    else:
        return a

def _transform(A, U):
    B = sla.solve_triangular(_conjugate(U), _conjugate(A), lower=True)
    A = sla.solve_triangular(_conjugate(U), _conjugate(B), lower=True)
    return A

def _default_block_size(left, right, extra, init, threads):
    extra_left = int(extra[0])
    extra_right = int(extra[1])
    init_left = 0
    init_right = 0
    if init[0] is not None:
        init_left = int(init[0].nvec())
    if init[1] is not None:
        init_right = int(init[1].nvec())
    if threads <= 8:
        threads = 8
    if left == 0 and right == 0:
        return 0
    if left <= 0 and right <= 0:
        if init_left == 0 and init_right == 0:
            if left < 0 and right < 0:
                return 2*threads
            else:
                return threads
        m = init_left + init_right
        m = threads*((m - 1)//threads + 1)
        if left < 0 or right < 0:
            m = max(m, 2*threads)
        return m
    left_total = 0
    right_total = 0
    if left > 0:
        if extra_left >= 0:
            left_total = max(left + extra_left, init_left)
        else:
            left_total = int(math.floor(max(left, init_left)*1.2))
    if right > 0:
        if extra_right >= 0:
            right_total = max(right + extra_right, init_right)
        else:
            right_total = int(math.floor(max(right, init_right)*1.2))
    if left < 0:
        left_total = right_total
    if right < 0:
        right_total = left_total
    m = int(left_total + right_total)
    m = threads*((m - 1)//threads + 1)
    if left < 0 or right < 0:
        m = max(m, 2*threads)
    return m

def _reset_cnv_data(i, cnv, res, acf, err_lmd, dlmd, err_X, dX, iterations):
    cnv[i] = 0
    res[i] = -1.0
    acf[:, i] = 1.0
    err_lmd[:, i] = -1.0
    dlmd[i, :] = 0
    err_X[:, i] = -1.0
    dX[i] = 1.0
    iterations[i] = 0

def _norm(a, axis):
    return numpy.apply_along_axis(nla.norm, axis, a)

def _piv_chol(A, k, eps, blk=64, verb=0):
    n = A.shape[0]
    buff = A[0, :].copy()
    ind = [i for i in range(n)]
    drop_case = 0
    dropped = 0
    last_check = -1
    if k > 0:
        U = sla.cholesky(A[:k, :k])
        A[:k, :k] = U.copy()
        A[:k, k : n] = sla.solve_triangular \
                       (_conjugate(U), A[:k, k : n], lower=True)
        A[k : n, :k].fill(0.0)
        A[k : n, k : n] -= numpy.dot(_conjugate(A[:k, k : n]), A[:k, k : n])
    l = k
    for i in range(k, n):
        s = numpy.diag(A[i : n, i : n]).copy()
        if i > l:
            t = _norm(A[l : i, i : n], 0)
            s -= t*t
        j = i + numpy.argmax(s)
        if i != j:
            buff[:] = A[i, :]
            A[i, :] = A[j, :]
            A[j, :] = buff
            buff[:] = A[:, i]
            A[:, i] = A[:, j]
            A[:, j] = buff
            ind[i], ind[j] = ind[j], ind[i]
        if i > l:
            A[i, i : n] -= numpy.dot(_conjugate(A[l : i, i]), A[l : i, i : n])
        last_piv = A[i, i].real
        if last_piv <= eps:
            A[i : n, :].fill(0.0)
            drop_case = 1
            dropped = n - i
            break
        A[i, i] = math.sqrt(last_piv)
        A[i, i + 1 : n] /= A[i, i]
        A[i + 1 : n, i].fill(0.0)
        if i - l == blk - 1 or i == n - 1:
            last_check = i
            lmin = _estimate_lmin(A[: i + 1, : i + 1])
            lmax = _estimate_lmax(A[: i + 1, : i + 1])
            if verb > 0:
                print('%e %e %e' % (A[i, i], lmin, lmax))
            if lmin/lmax <= eps:
                A[i : n, :].fill(0.0)
                drop_case = 2
                dropped = n - i
                break
        if i - l == blk - 1 and i < n - 1:
            j = i + 1
            A[j : n, j : n] -= numpy.dot(_conjugate(A[l : j, j : n]), \
                A[l : j, j : n])
            l += blk
    if last_check < n - 1 and drop_case != 2:
        i = last_check
        j = n - dropped - 1
        while i < j:
            m = i + (j - i + 1)//2
            lmin = _estimate_lmin(A[: m + 1, : m + 1])
            lmax = _estimate_lmax(A[: m + 1, : m + 1])
            if verb > 0:
                print('%d %e %e' % (m, lmin, lmax))
            if lmin/lmax <= eps:
                if j > m:
                    j = m
                    continue
                else:
                    A[j : n, :].fill(0.0)
                    dropped = n - j
                    last_piv = A[j - 1, j - 1]**2
                    break
            else:
                i = m
                continue
    return ind, dropped

def _estimate_lmax(U):
    U = numpy.triu(U)
    return sla.norm(numpy.dot(_conjugate(U), U), ord=1)
def _estimate_lmin(U):
    n = U.shape[0]
    if U.dtype.kind == 'c':
        tr = 2
    else:
        tr = 1
    x = numpy.ones((n,), dtype=U.dtype)
    s = numpy.dot(x, x)
    for i in range(3):
        y = sla.solve_triangular(U, x, trans=tr)
        t = numpy.dot(y, y)
        rq = s/t
        x = sla.solve_triangular(U, y)
        s = numpy.dot(x, x)
    return rq



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

# -*- coding: utf-8 -*-
'''Generates a data matrix from randomly generated singular vectors and 
   singular values of user-controlled behaviour.

   If run as a script, the generated matrix is saved to data.npy, and singular
   values and vectors to svd.npz.

   Alternatively, from generate_matrix import generate, after which the call of
   generate(m, n, rank, dtype, scale, alpha, pca) will return the matrix,
   singular values and left and right singular vectors.

Usage:
  generate_matrix [--help | -h | options] <m> <n> <r>

Arguments:
  m  number of rows in A
  n  number of columns in A
  r  rank of A (use r << min(m, n) for fast generation)

Options:
  -a <alpha>, --alpha=<alpha>  singular values decay rate [default: 0.75]
  -s <scale>, --scale=<scale>  singular values max value [default: 1.0]
  -d, --double   use double precision
  -p, --ptb      add random perturbation to make the matrix full rank
  -z, --pca      generate matrix with shift-to-zero-average invariant singular
                 values (save for the first), to be used for testing PCA

Note:
   Sigular values satisfy sigma_k ~ 1/k**alpha. With the default value of alfa,
   the singular spectrum of the randomly generated test matrix imitates that
   encountered in Principal Component Analyisis of images from lfw dataset
   (see http://vis-www.cs.umass.edu/lfw).

'''

import math
import numpy
import numpy.linalg as nla
import scipy.linalg as sla
import sys


def _norm(a, axis):
    return numpy.apply_along_axis(nla.norm, axis, a)


def random_singular_values(k, f_sigma, dt):
    s = numpy.random.rand(k).astype(dt)
    s = numpy.sort(s)
    s = f_sigma(s)
    return s/s[0]


def random_singular_vectors(m, n, k, dt, pca):
    u = numpy.random.randn(m, k).astype(dt)
    if pca:
        u[:, 0] = 1.0
    v = numpy.random.randn(n, k).astype(dt)
    u, r = sla.qr(u, mode='economic')
    v, r = sla.qr(v, mode='economic')
    return u, v


def random_matrix_for_svd(m, n, k, f_sigma, dt, pca=False):
    s = random_singular_values(min(m, n), f_sigma, dt)[:k]
    u, v = random_singular_vectors(m, n, k, dt, pca)
    a = numpy.dot(u*s, v.transpose())
    return s, u, v, a


def generate(m, n, rank, dtype=numpy.float32, scale=1.0, alpha=0.75, pca=False):
    f_sigma = lambda t: dtype(scale) * t**(-alpha)
    sigma, u, v, A = random_matrix_for_svd(m, n, rank, f_sigma, dtype, pca)
    return A, sigma, u, v


if __name__ == '__main__':
    try:
        from docopt import docopt
        __version__ = '0.1.0'
        have_docopt = True
    except:
        have_docopt = False

    if have_docopt:
        args = docopt(__doc__, version=__version__)
        m = int(args['<m>'])
        n = int(args['<n>'])
        k = int(args['<r>'])
        alpha = float(args['--alpha'])
        scale = float(args['--scale'])
        dble = args['--double']
        ptb = args['--ptb']
        pca = args['--pca']
    else:
        narg = len(sys.argv)
        if narg < 4:
            usage = 'Usage: generate_matrix <samples> <features> <rank>'
            raise SystemExit(usage)
        m = int(sys.argv[1])
        n = int(sys.argv[2])
        k = int(sys.argv[3])
        alpha = 0.75
        scale = 1.0
        dble = False
        ptb = False
        pca = False

    numpy.random.seed(1) # make results reproducible

    if dble:
        dtype = numpy.float64
    else:
        dtype = numpy.float32

    print('\n--- generating the matrix...')
    A, sigma, u, v = generate(m, n, k, dtype=dtype, scale=scale, alpha=alpha, \
                              pca=pca)
    if ptb:
        a = 2*numpy.random.rand(m, n).astype(dtype) - 1
        s = 10*_norm(a, axis=1)
        A += numpy.reshape(sigma[-1]/s, (m, 1))*a

    print('\n--- saving...')
    numpy.save('data.npy', A)
    numpy.savez('svd', sigma=sigma, left=u, right=v)

    print('done')



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

'''GPU implementation of RALEIGH dense algebra.
'''

import ctypes
import numbers
import numpy

from . import cuda_wrap as cuda
from . import cublas_handle as cublas
from .cublas_wrap import Cublas


class Vectors:
    '''CUBLAS implementation of Vectors type.
    '''

    '''========== Methods required by RALEIGH core solver ==========
    '''

    def new_vectors(self, arg=0, dim=None):
        try:
            nv = int(arg)
            if dim is None:
                dim = self.dimension()
            return Vectors(dim, nv, self.data_type()) #, cublas=self.__cublas)
        except:
            return Vectors(arg) #, cublas=self.__cublas)

    def clone(self):
        return Vectors(self)

    def append(self, other, axis=0):
        if other.nvec() < 1:
            return
        inc = ctypes.c_int(1)
        n = self.__vdim
        vsize = self.__dsize * n
        if axis == 1:
            m, n = self.shape()
            l, n_other = other.shape()
            if m != l:
                msg = 'Cannot append %d vectors to %d vectors' % (l, m)
                raise ValueError(msg)
            type_self = self.data_type()
            type_other = other.data_type()
            if type_self != type_other:
                rs = repr(type_self)
                ro = repr(type_other)
                msg = 'Cannot append %s vectors to %s vectors' % (ro, rs)
                raise ValueError(msg)
            vsize_other = self.__dsize * n_other
            n_new = n + n_other
            vsize_new = vsize + vsize_other
            vdata = _Data(m * vsize_new)
            ptr_n = vdata.data_ptr()
            ptr_s = self.data_ptr()
            ptr_o = other.data_ptr()
            ldn = ctypes.c_int(vsize_new)
            lds = ctypes.c_int(vsize)
            ldo = ctypes.c_int(vsize_other)
            h = ctypes.c_int(m)
            _try_calling(cuda.memcpy2D(ptr_n, ldn, ptr_s, lds, lds, h, \
                         cuda.memcpyD2D))
            ptr = _shifted_ptr(ptr_n, vsize)
            _try_calling(cuda.memcpy2D(ptr, ldn, ptr_o, ldo, ldo, h, \
                         cuda.memcpyD2D))
            self.__vdata = vdata
            self.__vdim = n_new
            return
        i, m = self.selected()
        j, l = other.selected()
        nvec = i + m + l
        if nvec > self.__mvec:
            mvec = ((nvec - 1)//self.__min_inc + 1)*self.__min_inc
#            mvec = ((nvec - 1)//Vectors.MIN_INC + 1)*Vectors.MIN_INC
#            print('allocating %d vectors...' % mvec)
            size = mvec * vsize
            vdata = _Data(size)
            data = vdata.data_ptr()
            if i + m > 0:
                mn = ctypes.c_int((i + m)*n)
                ptr_u = self.all_data_ptr()
                self.__cublas.copy(self.__cublas.handle, mn, ptr_u, inc, data, \
                                   inc)
            self.__vdata = vdata
            self.__mvec = mvec
            if self.__min_inc < self.__max_inc:
                self.__min_inc *= 2
        data = self.all_data_ptr()
        data_v = other.all_data_ptr()
        ptr_v = _shifted_ptr(data_v, j*vsize)
        ptr = _shifted_ptr(data, int(i + m)*vsize)
        ln = ctypes.c_int(l*n)
#        print('copying %d vectors...' % l)
        self.__cublas.copy(self.__cublas.handle, ln, ptr_v, inc, ptr, inc)
        self.__nvec = nvec
        self.select_all()

    def dimension(self):
        return self.__vdim

    def nvec(self):
        return self.__selected[1]

    def select(self, nv, first=0):
        assert nv <= self.__nvec and first >= 0
        self.__selected = (first, nv)

    def selected(self):
        return self.__selected

    def data_type(self):
        return self.__dtype

    def fill_random(self):
        n = self.dimension()
        m = self.nvec()
        if m < 1:
            return
        data = numpy.random.rand(m, n).astype(self.data_type())
        data *= 2
        data -= 1
        vsize = n * self.__dsize
        size = m * vsize
        ptr_u = self.data_ptr()
        ptr_v = ctypes.c_void_p(data.ctypes.data)
        _try_calling(cuda.memcpy(ptr_u, ptr_v, size, cuda.memcpyH2D))

    def copy(self, other, ind=None):
        i, m = self.selected()
        j, l = other.selected()
        vdim = self.dimension()
        inc = ctypes.c_int(1)
        vsize = self.__cublas.dsize * vdim
        data_u = self.all_data_ptr()
        data_v = other.all_data_ptr()
        if ind is None:
            n = ctypes.c_int(m*vdim)
            ptr_u = _shifted_ptr(data_u, i*vsize)
            ptr_v = _shifted_ptr(data_v, j*vsize)
            self.__cublas.copy(self.__cublas.handle, n, ptr_u, inc, ptr_v, inc)
        else:
            n = ctypes.c_int(vdim)
            l = len(ind)
            for k in range(l):
                ptr_u = _shifted_ptr(data_u, int(ind[k])*vsize)
                ptr_v = _shifted_ptr(data_v, (j + k)*vsize)
                self.__cublas.copy \
                    (self.__cublas.handle, n, ptr_u, inc, ptr_v, inc)

    def scale(self, s, multiply=False):
        f, m = self.selected()
        vdim = self.dimension()
        n = ctypes.c_int(vdim)
        inc = ctypes.c_int(1)
        vsize = self.__cublas.dsize * vdim
        data_u = self.all_data_ptr()
        if multiply:
            for i in range(m):
                ptr_u = _shifted_ptr(data_u, (f + i)*vsize)
                r = self.__to_floats(s[i])
                self.__cublas.scal(self.__cublas.handle, n, r, ptr_u, inc)
        else:
            for i in range(m):
                ptr_u = _shifted_ptr(data_u, (f + i)*vsize)
                if s[i] != 0.0:
                    r = self.__to_floats(1.0/s[i])
                    self.__cublas.scal(self.__cublas.handle, n, r, ptr_u, inc)

    def dots(self, other, transp=False):
        if transp:
            m = self.nvec()
            n = self.dimension()
            c_n = ctypes.c_int(n)
            c_m = ctypes.c_int(m)
            c_k = ctypes.c_int(1)
            d_pu = self.data_ptr()
            d_pv = other.data_ptr()
            size_dptr = ctypes.sizeof(d_pu)
            alpha = self.__to_floats(1.0)
            beta = self.__to_floats(0.0)
            size_w = n*self.__dsize
            w = numpy.ndarray((n,), dtype=self.data_type())
            h_pw = ctypes.c_void_p(w.ctypes.data)
            d_pw = Cublas.CtypesPtr()
            _try_calling(cuda.malloc(ctypes.byref(d_pw), size_w))
            iota = numpy.arange(n)*self.__dsize
            ptr_u = ctypes.cast(d_pu, ctypes.c_void_p)
            ptr_v = ctypes.cast(d_pv, ctypes.c_void_p)
            ptr_w = ctypes.cast(d_pw, ctypes.c_void_p)
            h_ppu = iota + ptr_u.value
            h_ppv = iota + ptr_v.value
            h_ppw = iota + ptr_w.value
            size_pp = n*size_dptr
            d_ppu = Cublas.CtypesPtr()
            d_ppv = Cublas.CtypesPtr()
            d_ppw = Cublas.CtypesPtr()
            _try_calling(cuda.malloc(ctypes.byref(d_ppu), size_pp))
            _try_calling(cuda.malloc(ctypes.byref(d_ppv), size_pp))
            _try_calling(cuda.malloc(ctypes.byref(d_ppw), size_pp))
            _try_calling(cuda.memcpy(d_ppu, h_ppu.ctypes.data, size_pp, cuda.memcpyH2D))
            _try_calling(cuda.memcpy(d_ppv, h_ppv.ctypes.data, size_pp, cuda.memcpyH2D))
            _try_calling(cuda.memcpy(d_ppw, h_ppw.ctypes.data, size_pp, cuda.memcpyH2D))
            if self.__is_complex:
                Trans = Cublas.ConjTrans
            else:
                Trans = Cublas.Trans
            _try_calling(self.__cublas.gemm_batched(self.__cublas.handle, \
                Cublas.NoTrans, Trans, \
                c_k, c_k, c_m, alpha, d_ppu, c_n, d_ppv, c_n, \
                beta, d_ppw, c_k, c_n))
            _try_calling(cuda.memcpy(h_pw, d_pw, size_w, cuda.memcpyD2H))
            _try_calling(cuda.free(d_ppu))
            _try_calling(cuda.free(d_ppv))
            _try_calling(cuda.free(d_ppw))
            _try_calling(cuda.free(d_pw))
            return w
        iu = self.first()
        iv = other.first()
        vdim = self.dimension()
        dsize = self.__dsize
        vsize = dsize * vdim
        m = self.nvec()
        n = ctypes.c_int(vdim)
        inc = ctypes.c_int(1)
        data_u = self.all_data_ptr()
        data_v = other.all_data_ptr()
        w = numpy.ndarray((m,), dtype=self.data_type())
        for i in range(m):
            ptr_u = _shifted_ptr(data_u, (iu + i)*vsize)
            ptr_v = _shifted_ptr(data_v, (iv + i)*vsize)
            s = self.__floats()
            self.__cublas.dot \
                (self.__cublas.handle, n, ptr_v, inc, ptr_u, inc, s)
            if self.is_complex():
                w[i] = s[0] + 1j * s[1]
            else:
                w[i] = s[0]
        return w

    def dot(self, other):
        m = self.nvec()
        n = self.dimension()
        k = other.nvec()
        c_n = ctypes.c_int(n)
        c_m = ctypes.c_int(m)
        c_k = ctypes.c_int(k)
        dptr_u = other.data_ptr()
        dptr_v = self.data_ptr()
        dptr_q = ctypes.POINTER(ctypes.c_ubyte)()
        size_q = int(k*m*self.__dsize)
        _try_calling(cuda.malloc(ctypes.byref(dptr_q), size_q))
        if self.is_complex():
            Trans = Cublas.ConjTrans
        else:
            Trans = Cublas.Trans
        one = self.__to_floats(1.0)
        zero = self.__to_floats(0.0)
        self.__cublas.gemm(self.__cublas.handle, Trans, Cublas.NoTrans, \
            c_m, c_k, c_n, one, dptr_v, c_n, dptr_u, c_n, zero, dptr_q, c_m)
        q = numpy.ndarray((k, m), dtype=self.data_type())
        hptr_q = ctypes.c_void_p(q.ctypes.data)
        _try_calling(cuda.memcpy(hptr_q, dptr_q, size_q, cuda.memcpyD2H))
        _try_calling(cuda.free(dptr_q))
        return _conjugate(q)

    def multiply(self, a, output):
        m = a.shape[1]
        n = self.dimension()
        k = self.nvec()
        c_n = ctypes.c_int(n)
        c_m = ctypes.c_int(m)
        c_k = ctypes.c_int(k)
        dptr_u = output.data_ptr()
        dptr_v = self.data_ptr()
        dptr_q = ctypes.POINTER(ctypes.c_ubyte)()
        size_q = int(k*m*self.__dsize)
        _try_calling(cuda.malloc(ctypes.byref(dptr_q), size_q))
        if a.flags['C_CONTIGUOUS'] or a.flags['F_CONTIGUOUS']:
            q = a
        else:
            q = a.copy()
        hptr_q = ctypes.c_void_p(q.ctypes.data)
        _try_calling(cuda.memcpy(dptr_q, hptr_q, size_q, cuda.memcpyH2D))
        if q.flags['C_CONTIGUOUS']:
            Trans = Cublas.Trans
            ldq = c_m
        elif q.flags['F_CONTIGUOUS']:
            Trans = Cublas.NoTrans
            ldq = c_k
        one = self.__to_floats(1.0)
        zero = self.__to_floats(0.0)
        self.__cublas.gemm(self.__cublas.handle, Cublas.NoTrans, Trans, \
            c_n, c_m, c_k, one, dptr_v, c_n, dptr_q, ldq, zero, dptr_u, c_n)
        _try_calling(cuda.free(dptr_q))

    def add(self, other, s, a=None):
        n = self.dimension()
        m = other.nvec()
        c_n = ctypes.c_int(n)
        c_m = ctypes.c_int(m)
        vsize = self.__dsize * n
        dptr_u = other.data_ptr()
        dptr_v = self.data_ptr()
        if numpy.isscalar(s):
            cublas_s = self.__to_floats(s)
            if a is None:
                nm = ctypes.c_int(n*m)
                inc = ctypes.c_int(1)
                self.__cublas.axpy \
                    (self.__cublas.handle, nm, cublas_s, dptr_u, inc, dptr_v, \
                     inc)
            else:
                k = a.shape[1]
                c_k = ctypes.c_int(k)
                if a.flags['C_CONTIGUOUS']:
                    Trans = Cublas.Trans
                    q = a
                    ldq = c_k
                elif a.flags['F_CONTIGUOUS']:
                    Trans = Cublas.NoTrans
                    q = a
                    ldq = c_m
                else:
                    q = numpy.ndarray(a.shape, dtype=a.dtype)
                    q[:,:] = a.copy()
                    Trans = Cublas.Trans
                    ldq = c_k
                hptr_q = ctypes.c_void_p(q.ctypes.data)
                dptr_q = ctypes.POINTER(ctypes.c_ubyte)()
                size_q = k*m*self.__dsize
                _try_calling(cuda.malloc(ctypes.byref(dptr_q), size_q))
                _try_calling(cuda.memcpy(dptr_q, hptr_q, size_q, cuda.memcpyH2D))
                one = self.__to_floats(1.0)
                self.__cublas.gemm(self.__cublas.handle, Cublas.NoTrans, Trans, \
                    c_n, c_k, c_m, cublas_s, dptr_u, c_n, dptr_q, ldq, \
                    one, dptr_v, c_n)
                _try_calling(cuda.free(dptr_q))
        else:
            for i in range(m):
                dptr_u = _shifted_ptr(other.data_ptr(), i*vsize)
                dptr_v = _shifted_ptr(self.data_ptr(), i*vsize)
                inc = ctypes.c_int(1)
                cublas_s = self.__to_floats(s[i])
                self.__cublas.axpy(self.__cublas.handle, c_n, cublas_s, \
                                   dptr_u, inc, dptr_v, inc)

    '''========== Other methods ====================================
    '''

    def __init__(self, arg, nvec=0, data_type=None, shallow=False): #, cublas=None):
        if isinstance(arg, Vectors):
            n = arg.dimension()
            m = arg.nvec()
            self.__is_complex = arg.__is_complex
            dtype = arg.data_type()
            dsize = arg.data_size()
            if shallow:
                self.__vdata = arg.vectors_data()
            else:
                size = n*m*dsize
                self.__vdata = _Data(size)
                _try_calling(cuda.memcpy(self.all_data_ptr(), arg.data_ptr(), \
                                         size, cuda.memcpyD2D))
        elif isinstance(arg, Matrix):
            if arg.order() is not 'C_CONTIGUOUS':
                raise ValueError('Vectors data must be C_CONTIGUOUS')
            m, n = arg.shape()
            dtype = arg.data_type()
            dsize = arg.data_size()
            self.__is_complex = arg.is_complex()
            self.__vdata = arg.matrix_data() #TODO: deep copy case
        elif isinstance(arg, numpy.ndarray):
            m, n = arg.shape
            dtype = arg.dtype.type
            dsize = arg.itemsize
            size = n*m*dsize
            self.__vdata = _Data(size)
            ptr = ctypes.c_void_p(arg.ctypes.data)
            _try_calling(cuda.memcpy(self.all_data_ptr(), ptr, size, \
                                     cuda.memcpyH2D))
            self.__is_complex = \
                (dtype == numpy.complex64 or dtype == numpy.complex128)
        elif isinstance(arg, numbers.Number):
            dtype = data_type
            if dtype is None:
                dtype = numpy.float64
            if dtype == numpy.float32:
                dsize = 4
                self.__is_complex = False
            elif dtype == numpy.float64:
                dsize = 8
                self.__is_complex = False
            elif dtype == numpy.complex64:
                dsize = 8
                self.__is_complex = True
            elif dtype == numpy.complex128:
                dsize = 16
                self.__is_complex = True
            else:
                raise ValueError('data type %s not supported' % repr(dtype))
            n = arg
            m = nvec
            assert m >= 0
            if nvec > 0:
                size = int(n*m*dsize)
                self.__vdata = _Data(size)
                _try_calling(cuda.memset(self.all_data_ptr(), 0, size))
            else:
                self.__vdata = None
        else:
            raise ValueError \
                ('wrong argument %s in constructor' % repr(type(arg)))
        if cublas.handle is None:
            cublas.handle = Cublas(dtype)
        self.__selected = (0, m)
        self.__vdim = n
        self.__nvec = m
        self.__mvec = m
        self.__min_inc = 16
        self.__max_inc = 1024
        self.__dsize = dsize
        self.__dtype = dtype
        self.__cublas = cublas.handle

    def __float(self):
        dt = self.data_type()
        if dt == numpy.float32:
            return ctypes.c_float()
        elif dt == numpy.float64:
            return ctypes.c_double()
        else:
            raise ValueError('wrong data type %s passed to __float' % repr(dt))

    def __floats(self):
        dt = self.data_type()
        if dt == numpy.float32:
            floats = ctypes.c_float * 1
        elif dt == numpy.float64:
            floats = ctypes.c_double * 1
        elif dt == numpy.complex64:
            floats = ctypes.c_float * 2
        elif dt == numpy.complex128:
            floats = ctypes.c_double * 2
        else:
            raise ValueError('wrong data type %s passed to __complex' % repr(dt))
        return floats()

    def __to_floats(self, v):
        s = self.__floats()
        dt = self.data_type()
        if dt == numpy.float32:
            s[0] = ctypes.c_float(v)
        elif dt == numpy.float64:
            s[0] = ctypes.c_double(v)
        elif dt == numpy.complex64 or dt == numpy.complex128:
            s[0] = v.real
            s[1] = v.imag
        else:
            raise ValueError('data type %s not supported' % repr(dt))
        return s

    def shape(self):
        return (self.__nvec, self.__vdim)

    def all_data_ptr(self):
        return self.__vdata.data_ptr()

    def data_ptr(self):
        n = self.dimension()
        vsize = n * self.__dsize
        return _shifted_ptr(self.all_data_ptr(), self.first() * vsize)
        
    def vectors_data(self):
        return self.__vdata

    def data_size(self):
        return self.__dsize

    def cublas(self):
        return self.__cublas

    def cublas_handle(self):
        return self.__cublas.handle

    def first(self):
        return self.__selected[0]

    def select_all(self):
        self.select(self.__nvec)

    def reference(self):
        v = Vectors(self, shallow=True)
        return v

    def is_complex(self):
        return self.__is_complex

    def conjugate(self):
        if not self.is_complex():
            return
        s = self.__to_floats(-1.0)
        m = self.nvec()
        n = self.dimension()
        inc = ctypes.c_int(2)
        ptr = _shifted_ptr(self.data_ptr(), self.data_size()//2)
        self.cublas().fscal(self.cublas().handle, m*n, s, ptr, inc)

    def orthogonalize(self, other):
        m = self.nvec()
        n = self.dimension()
        k = other.nvec()
        q = self.new_vectors(k, m)
        c_n = ctypes.c_int(n)
        c_m = ctypes.c_int(m)
        c_k = ctypes.c_int(k)
        dptr_u = other.data_ptr()
        dptr_v = self.data_ptr()
        dptr_q = q.data_ptr()
        if self.is_complex():
            Trans = Cublas.ConjTrans
        else:
            Trans = Cublas.Trans
        zero = self.__to_floats(0.0)
        pl_one = self.__to_floats(1.0)
        mn_one = self.__to_floats(-1.0)
        self.__cublas.gemm(self.__cublas.handle, Trans, Cublas.NoTrans, \
            c_m, c_k, c_n, pl_one, dptr_v, c_n, dptr_u, c_n, zero, dptr_q, c_m)
        self.__cublas.gemm(self.__cublas.handle, Cublas.NoTrans, Trans, \
            c_n, c_m, c_k, mn_one, dptr_u, c_n, dptr_q, c_m, pl_one, dptr_v, c_n)
        return q

    def svd(self):
        try:
            from .dense_cblas import Vectors as cpuVectors
            #print('using mkl cblas...')
        except:
            #print('mkl cblas not found, using numpy...')
            from .dense_numpy import Vectors as cpuVectors
        u = cpuVectors(self.data())
        sigma, vt = u.svd()
        self.fill(u.data())
        return sigma, vt
#        m = self.nvec()
#        n = self.dimension()
#        c_n = ctypes.c_int(n)
#        c_m = ctypes.c_int(m)
#        c_lw = ctypes.c_int()
#        err = self.__cublas.svd_buff_size(self.__cublas.cusolver_handle, \
#            c_n, c_m, ctypes.byref(c_lw))
#        lw = c_lw.value
#        print('buffer size: %d' % lw)
#        dptr_r = Cublas.CtypesPtr()
#        dptr_s = Cublas.CtypesPtr()
#        dptr_u = self.data_ptr()
#        dptr_v = Cublas.CtypesPtr()
#        dptr_w = Cublas.CtypesPtr()
#        dptr_i = Cublas.CtypesPtr()
#        dtype = self.data_type()
#        dsize = self.data_size()
#        _try_calling(cuda.malloc(ctypes.byref(dptr_w), lw*dsize))
#        v = numpy.ndarray((m, m), dtype=dtype)
#        _try_calling(cuda.malloc(ctypes.byref(dptr_v), m*m*dsize))
#        if dtype == numpy.complex64:
#            dtype = numpy.float32
#            dsize = 4
#        elif dtype == numpy.complex128:
#            dtype = numpy.float64            
#            dsize = 8
#        sigma = numpy.ndarray((m,), dtype=dtype)
#        _try_calling(cuda.malloc(ctypes.byref(dptr_r), m*dsize))
#        _try_calling(cuda.malloc(ctypes.byref(dptr_s), m*dsize))
#        _try_calling(cuda.malloc(ctypes.byref(dptr_i), 4))
#        jobu = self.__cublas.ovwrt
#        jobv = self.__cublas.all
#        err = self.__cublas.svd(self.__cublas.cusolver_handle, jobu, jobv, \
#            n, m, dptr_u, n, dptr_s, dptr_u, n, dptr_v, m, \
#            dptr_w, lw, dptr_r, dptr_i)
##            c_n, c_m, dptr_u, c_n, dptr_s, dptr_u, c_n, dptr_v, c_m, \
##            dptr_w, c_lw, dptr_r, dptr_i)
#        hptr_s = ctypes.c_void_p(sigma.ctypes.data)
#        _try_calling(cuda.memcpy(hptr_s, dptr_s, m*dsize, cuda.memcpyD2H))
#        i = ctypes.c_int()
#        _try_calling(cuda.memcpy(ctypes.byref(i), dptr_i, 4, cuda.memcpyD2H))
#        _try_calling(cuda.free(dptr_r))
#        _try_calling(cuda.free(dptr_s))
#        _try_calling(cuda.free(dptr_v))
#        _try_calling(cuda.free(dptr_w))
#        _try_calling(cuda.free(dptr_i))
#        return sigma, v.T

    def zero(self):
        m = self.nvec()
        if m < 1:
            return
        vsize = self.__dsize * self.__vdim
        _try_calling(cuda.memset(self.data_ptr(), 0, m*vsize))

    def fill(self, data):
        m, n = data.shape
        if m != self.__nvec or n != self.__vdim:
            raise ValueError('mismatching dimensions in fill()')
        if m < 1:
            return
        dtype = data.dtype.type
        if dtype != self.__dtype:
            raise ValueError('mismatching data types in fill()')
        size = m * self.__dsize * self.__vdim
        ptr = ctypes.c_void_p(data.ctypes.data)
        _try_calling(cuda.memcpy(self.all_data_ptr(), ptr, size, cuda.memcpyH2D))

    def data(self):
        m = self.nvec()
        n = self.dimension()
        v = numpy.ndarray((m, n), dtype=self.data_type())
        if m < 1:
            return v
        hptr_v = ctypes.c_void_p(v.ctypes.data)
        size = int(n * m * self.__dsize)
        _try_calling(cuda.memcpy(hptr_v, self.data_ptr(), size, cuda.memcpyD2H))
        return v

    def asarray(self):
        return self.data().T


class Matrix:

    def __init__(self, arg):
        if isinstance(arg, Vectors):
            n = arg.dimension()
            m = arg.nvec()
            self.__shape = (m, n)
            self.__dtype = arg.data_type()
            self.__dsize = arg.data_size()
            self.__is_complex = arg.is_complex()
            self.__order = 'C_CONTIGUOUS'
            self.__mdata = arg.vectors_data()
        elif isinstance(arg, numpy.ndarray):
            self.__shape = arg.shape
            self.__dtype = arg.dtype.type
            self.__dsize = arg.itemsize
            self.__is_complex = (arg.dtype.kind == 'c')
            if arg.flags['C_CONTIGUOUS']:
                self.__order = 'C_CONTIGUOUS'
            elif arg.flags['F_CONTIGUOUS']:
                self.__order = 'F_CONTIGUOUS'
            else:
                msg = 'Matrix data must be either C- or F-contiguous'
                raise ValueError(msg)
            m, n = self.__shape
            size = n*m*self.__dsize
            self.__mdata = _Data(size)
            ptr = ctypes.c_void_p(arg.ctypes.data)
            _try_calling(cuda.memcpy(self.data_ptr(), ptr, size, cuda.memcpyH2D))
        else:
            raise ValueError \
                ('wrong argument %s in Matrix constructor' % repr(type(arg)))

    def __floats(self):
        dt = self.__dtype
        if dt == numpy.float32:
            floats = ctypes.c_float * 1
        elif dt == numpy.float64:
            floats = ctypes.c_double * 1
        elif dt == numpy.complex64:
            floats = ctypes.c_float * 2
        elif dt == numpy.complex128:
            floats = ctypes.c_double * 2
        else:
            raise ValueError('wrong data type %s passed to __complex' % repr(dt))
        return floats()

    def __to_floats(self, v):
        s = self.__floats()
        dt = self.__dtype
        if dt == numpy.float32:
            s[0] = ctypes.c_float(v)
        elif dt == numpy.float64:
            s[0] = ctypes.c_double(v)
        elif dt == numpy.complex64 or dt == numpy.complex128:
            s[0] = v.real
            s[1] = v.imag
        else:
            raise ValueError('data type %s not supported' % repr(dt))
        return s
        
    def data_ptr(self):
        return self.__mdata.data_ptr()

    def matrix_data(self):
        return self.__mdata

    def order(self):
        return self.__order

    def shape(self):
        return self.__shape

    def data_type(self):
        return self.__dtype

    def data_size(self):
        return self.__dsize

    def is_complex(self):
        return self.__is_complex

    def fill(self, data):
        m, n = self.__shape
        size = n*m*self.__dsize
        ptr = ctypes.c_void_p(data.ctypes.data)
        _try_calling(cuda.memcpy(self.data_ptr(), ptr, size, cuda.memcpyH2D))

    def dots(self):
        v = Vectors(self, shallow=True)
        return v.dots(v)

    def new_vectors(self, dim=None, nv=0):
        if dim is None:
            dim = self.shape()[1]
        return Vectors(dim, nv, self.data_type())

    def apply(self, x, y, transp=False):
        if x.data_type() != self.__dtype or y.data_type() != self.__dtype:
            raise ValueError('Matrix and vectors data types differ')
        m, n = self.__shape
        if transp:
            if n != y.dimension() or m != x.dimension():
                raise ValueError('Matrix and vectors dimensions incompatible')
        else:
            if m != y.dimension() or n != x.dimension():
                raise ValueError('Matrix and vectors dimensions incompatible')
        k = x.nvec()
        if k != y.nvec():
            raise ValueError('Numbers of input and output vectors differ')
        k = ctypes.c_int(k)
        nx = ctypes.c_int(x.dimension())
        ny = ctypes.c_int(y.dimension())
        conj = False
        if self.__order == 'C_CONTIGUOUS':
            if transp:
                if self.__is_complex:
                    x.conjugate()
                    conj = True
                Trans = Cublas.NoTrans
            else:
                Trans = Cublas.Trans
            lda = ctypes.c_int(n)
        elif self.__order == 'F_CONTIGUOUS':
            if transp:
                if self.__is_complex:
                    Trans = Cublas.ConjTrans
                else:
                    Trans = Cublas.Trans
            else:
                Trans = Cublas.NoTrans
            lda = ctypes.c_int(m)
        one = self.__to_floats(1.0)
        zero = self.__to_floats(0.0)
        dptr_a = self.data_ptr()
        dptr_x = x.data_ptr()
        dptr_y = y.data_ptr()
        x.cublas().gemm(x.cublas_handle(), Trans, Cublas.NoTrans, \
            ny, k, nx, one, dptr_a, lda, dptr_x, nx, zero, dptr_y, ny)
        if conj:
            x.conjugate()
            y.conjugate()


def _try_calling(err):
    if err != 0:
        raise RuntimeError('cuda error %d' % err)


def _shifted_ptr(dev_ptr, shift=0):
    ptr = ctypes.cast(dev_ptr, ctypes.c_void_p)
    return ctypes.cast(ptr.value + shift, ctypes.POINTER(ctypes.c_ubyte))


def _ptr_value(dev_ptr):
    ptr = ctypes.cast(dev_ptr, ctypes.c_void_p)
    return ptr.value


def _conjugate(a):
    if a.dtype.kind == 'c':
        return a.conj()
    else:
        return a


class _Data:

    def __init__(self, size):
        self.__data = ctypes.POINTER(ctypes.c_ubyte)()
        _try_calling(cuda.malloc(ctypes.byref(self.__data), size))

    def __del__(self):
        _try_calling(cuda.free(self.__data))

    def data_ptr(self):
        return self.__data



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

"""Pure numpy implementation of RALEIGH dense algebra.
"""

import numpy

from .dense_ndarray import NDArrayVectors, NDArrayMatrix


class Vectors(NDArrayVectors):
    '''Pure numpy implementation of ndarray-based Vectors type.
    '''

    '''========== Methods required by RALEIGH core solver ==========
    '''

    def new_vectors(self, arg=0, dim=None):
        if isinstance(arg, numpy.ndarray):
            return Vectors(arg)
        nv = arg
        if dim is None:
            dim = self.dimension()
        return Vectors(dim, nv, self.data_type())

#    def new_vectors(self, nv=0, dim=None):
#        if dim is None:
#            dim = self.dimension()
#        return Vectors(dim, nv, self.data_type())

    def clone(self):
        return Vectors(self)

    def copy(self, other, ind=None):
        i, n = self.selected()
        j, m = other.selected()
        if ind is None:
            assert(m == n)
            other.data()[:,:] = self.data()
        else:
            other.all_data()[j : j + len(ind), :] = self.all_data()[ind, :]

    def scale(self, s, multiply=False):
        f, n = self.selected();
        if multiply:
            for i in range(n):
                self.data(i)[:] *= s[i]
        else:
            for i in range(n):
                if s[i] != 0.0:
                    self.data(i)[:] /= s[i]

    def dots(self, other, transp=False):
        if transp:
            u = self.data()
            v = other.data()
            n = self.dimension()
            w = numpy.ndarray((n,), dtype=self.data_type())
            if other.is_complex():
                for i in range(n):
                    w[i] = numpy.dot(v[:, i].conj(), u[:, i])
            else:
                for i in range(n):
                    w[i] = numpy.dot(v[:, i], u[:, i])
            return w
        else:
            n = self.nvec()
        v = numpy.ndarray((n,), dtype=self.data_type())
        for i in range(n):
            if other.is_complex():
                s = numpy.dot(other.data(i).conj(), self.data(i))
            else:
                s = numpy.dot(other.data(i), self.data(i))
            v[i] = s
        return v

    def dot(self, other):
        if other.is_complex():
            return numpy.dot(other.data().conj(), self.data().T)
        else:
            return numpy.dot(other.data(), self.data().T)

    def multiply(self, q, output):
        f, s = output.selected();
        m = q.shape[1]
        assert(s == m)
        if output.data().flags['C_CONTIGUOUS']:
            #print('using optimized dot')
            numpy.dot(q.T, self.data(), out=output.data())
            return
        print('using non-optimized dot')
        output.data()[:,:] = numpy.dot(q.T, self.data())

    def add(self, other, s, q = None):
        f, m = self.selected();
        if numpy.isscalar(s):
            if q is None:
                self.data()[:,:] += s*other.data()
            else:
                self.data()[:,:] += s*numpy.dot(q.T, other.data())
            return
        else:
            for i in range(m):
                self.data(i)[:] += s[i]*other.data(i)

    '''========== Other methods ====================================
    '''

    def __init__(self, arg, nvec=0, data_type=None, shallow=False):
        super(Vectors, self).__init__(arg, nvec, data_type, shallow)
        
    def reference(self):
        v = Vectors(self, shallow=True)
        return v

    def orthogonalize(self, other):
        if other.is_complex():
            q = numpy.dot(other.data().conj(), self.data().T)
        else:
            q = numpy.dot(other.data(), self.data().T)
        self.data()[:,:] -= numpy.dot(q.T, other.data())
        return self.new_vectors(q)

    def svd(self):
        v, sigma, self.data()[:,:] = \
            numpy.linalg.svd(self.data(), full_matrices=False)
        return sigma, v.T

    def apply(self, A, output, transp=False):
        a = A.data()
        if transp:
            is_complex = A.is_complex()
            if is_complex:
                numpy.conj(self.data(), out=self.data())
            self.__apply(a.T, output)
            if is_complex:
                numpy.conj(output.data(), out=output.data())
                numpy.conj(self.data(), out=self.data())
        else:
            self.__apply(a, output)
    def __apply(self, q, output):
        if output.data().flags['C_CONTIGUOUS']:
            #print('using optimized dot')
            numpy.dot(self.data(), q.T, out=output.data())
        else:
            #print('using non-optimized dot')
            output.data()[:,:] = numpy.dot(self.data(), q.T)


class Matrix(NDArrayMatrix):

    def apply(self, x, y, transp=False):
        if transp:
            is_complex = self.is_complex()
            if is_complex:
                numpy.conj(x.data(), out=x.data())
            self.__apply(x, y, transp)
            if is_complex:
                numpy.conj(x.data(), out=x.data())
                numpy.conj(y.data(), out=y.data())
        else:
            self.__apply(x, y)

    def __apply(self, x, y, transp=False):
        if transp:
            a = self.data().T
        else:
            a = self.data()
        if y.data().flags['C_CONTIGUOUS']:
            #print('using optimized dot')
            numpy.dot(x.data(), a.T, out=y.data())
        else:
            print('using non-optimized dot')
            y.data()[:,:] = numpy.dot(x.data(), a.T)

    def dots(self):
        v = Vectors(self, shallow=True)
        return v.dots(v)

    def new_vectors(self, dim=None, nv=0):
        if dim is None:
            dim = self.shape()[1]
        return Vectors(dim, nv, self.data_type())


# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

# -*- coding: utf-8 -*-
"""LowerRankApproximation of a matrix.
"""

import math
import numpy
import numpy.linalg as nla
import scipy.linalg as sla

from ..core.solver import Options
from ..algebra.dense_matrix import AMatrix
from .partial_svd import PartialSVD
from .truncated_svd import DefaultStoppingCriteria


class LowerRankApproximation:
    '''Class for handling the computation of a lower rank approximation of
    a dense matrix, see method compute below for details.
    '''
    def __init__(self, mlr=None):
        if mlr is None:
            self.__left = None
            self.__right = None
            self.__mean = None
            self.__rank = 0
            self.__dtype = None
        else:
            self.__mean = mlr[0]
            self.__left = mlr[1]
            self.__right = mlr[2]
            self.__rank = self.__right.shape[0]
            self.__dtype = self.__left.dtype.type
        self.__left_v = None
        self.__right_v = None
        self.__mean_v = None
        self.__tol = -1
        self.__svtol = 1e-3
        self.__norm = None
        self.__arch = None
        self.ortho = True
        self.iterations = -1
        
    def compute(self, data_matrix, opt=Options(), rank=-1, tol=0, norm='f',
                max_rank=-1, svtol=1e-3, shift=False, verb=0):
        '''
        For a given m by n data matrix A (m data samples n features each)
        computes m by k matrix L and k by n matrix R such that k <= min(m, n),
        and the product L R approximates A if shift is False or else A - e a,
        where e = numpy.ones((m, 1)) and a = numpy.mean(A, axis=0).

        The rows of R are orhonormal, the columns of L are in the descending
        order of their norms.

        Below A_ stands for A if shift is False and for A - e a otherwise.

        Parameters
        ----------
        data_matrix : AMatrix
            Data matrix A.
        opt : an object of class raleigh.solver.Options
            Solver options (see raleigh.solver).
        rank : int
            Required number of columns in L = number of rows in R (k in the
            above description of the method).
            If negative, implicitely defined by the required accuracy of
            approximation or interactively by the user.
        tol : float
            Approximation tolerance in the case rank < 0. If tol is non-zero, 
            then the computation of the lower rank approximation will stop
            when the norm of the difference D = A_ - L R becomes not greater
            than eps, where eps is the norm of A multiplied by tol if tol > 0
            and eps = -tol if tol < 0. Otherwise the user will be asked 
            repeatedly whether the computation should continue (the number of
            computed principal components and the relative truncation error, 
            the ratio of the norm of D to that of A, are displayed).
        norm : character
            The norm to be used for evaluating the approximation error:
            's' : the largest singular value of D (if tol > 0, then the largest
                  singular value of A - e a will be used instead of the norm of
                  A for evaluating the relative truncation error),
            'f' : Frobenius norm of D,
            'm' : the largest norm of a row of D.
        max_rank : int
            Maximal acceptable rank of L and R. Ignored if negatide, otherwise
            if max_rank < min(m, n), then the required accuracy of approximation
            might not be achieved.
        svtol : float
            Error tolerance for singular values (see Notes below) relative to
            the largest singular value.
        shift : bool
            Specifies whether L R approximates A (shift=False) or A_ = A - e a
            (shift=True, see the above description of the method).
        verb : int
            Verbosity level.

        Notes
        -----
        The rows of R are approximate right singular vectors of A_.
        The columns of L are approximate left singular vectors of A_ multiplied
        by respective singular values.
        Singular values and vectors are computed by applying block
        Jacobi-Conjugated Gradient algorithm to A_.T A_ or A_ A_.T, whichever
        is smaller in size.
        '''
        psvd = PartialSVD(data_matrix, shift)
        m, n = data_matrix.shape()

        user_bs = opt.block_size
        if user_bs < 1 and (rank < 0 or rank > 100):
            opt.block_size = 128
        if opt.convergence_criteria is None:
            no_cc = True
            opt.convergence_criteria = _DefaultLRAConvergenceCriteria(svtol)
        else:
            no_cc = False
        if opt.stopping_criteria is None and rank < 0:
            no_sc = True
            opt.stopping_criteria = \
                DefaultStoppingCriteria(data_matrix, tol, norm, max_rank, verb)
            v = psvd.vectors()
            opSVD = psvd.op_svd()
            opt.stopping_criteria.err_calc.set_up(opSVD, v, shift)
        else:
            no_sc = False

        psvd.compute(data_matrix, opt=opt, nsv=(0, rank), \
            refine=self.ortho)
        self.__left_v = psvd.left_v()
        self.__left_v.scale(psvd.sigma, multiply=True)
        self.__right_v = psvd.right_v()
        self.__mean_v = psvd.mean_v()
        self.__rank = self.__left_v.nvec()
        self.__opt = opt
        self.__tol = tol
        self.__svtol = svtol
        self.__norm = norm
        self.__arch = data_matrix.arch()
        self.__dtype = data_matrix.data_type()
        if max_rank > 0 and self.__left_v.nvec() > max_rank:
            self.__left_v.select(max_rank)
            self.__right_v.select(max_rank)
        self.iterations = psvd.iterations
        
        # restore user opt to avoid side effects
        opt.block_size = user_bs
        if no_cc:
            opt.convergence_criteria = None
        if no_sc:
            opt.stopping_criteria = None

    def update(self, data_matrix, opt=None, rank=-1, max_rank=-1, \
            tol=None, norm=None, svtol=None, verb=0):
        '''
        Updates previously computed Lower Rank Approximation of a data matrix
        data_matrix0, i.e. computes Lower Rank Approximation of 
        numpy.concatenate((data_matrix0, data_matrix))

        Parameters have the same meaning as for compute().
        '''
        if self.__rank == 0:
            raise RuntimeError('no existing LRA data to update')
        if opt is None:
            opt = self.__opt
        if tol is None:
            tol = self.__tol
        if tol == 0.0 and rank < 1:
            rank = self.__rank
        if norm is None:
            norm = self.__norm
        if svtol is None:
            svtol = self.__svtol
        if norm not in ['f', 'm', 's']:
            msg = 'norm %s is not supported' % repr(norm)
            raise ValueError(msg)
        v = data_matrix.as_vectors()
        s = abs(v.dots(v))
        fnorm = math.sqrt(numpy.sum(s))
        maxl2norm = numpy.amax(numpy.sqrt(s))
        if maxl2norm == 0.0:
            return
        dtype = self.__dtype
        if self.__left_v is None:
            left_data = self.__left.T
            # TODO: this must be taken care of by new_vectors()
            if not left_data.flags['C_CONTIGUOUS']:
                left_data = numpy.ndarray(left_data.shape, dtype=dtype)
                left_data[:,:] = self.__left.T.copy()
            self.__left_v = v.new_vectors(left_data)
            self.__right_v = v.new_vectors(self.__right)
            if self.__mean is not None:
                self.__mean_v = v.new_vectors(self.__mean)
            else:
                self.__mean_v = None
            self.__arch = data_matrix.arch()
        else:
            if self.__arch != data_matrix.arch() or \
                dtype != data_matrix.data_type():
                raise ValueError('incompatible matrix type passed to update')
        left0 = self.__left_v
        right0 = self.__right_v
        if self.ortho is False:
            wl = left0.new_vectors(left0.nvec())
            wr = right0.new_vectors(right0.nvec())
            H = right0.dot(right0)
            mu, x = sla.eigh(H)
            q = mu[0]/mu[-1]
            print(mu[0], mu[-1])
            if q < 0.5:
                _lra_ortho(left0, right0, wl, wr)
            else:
                G = left0.dot(left0)
                lmd, x = sla.eigh(-G, H)
                y = nla.inv(x.T)
                left0.multiply(y, wl)
                wl.copy(left0)
                right0.multiply(x, wr)
                wr.copy(right0)
        shift = self.__mean_v is not None
        sigma = numpy.sqrt(left0.dots(left0))
        sigma0 = sigma[0]
        n0 = left0.dimension()
        e0 = numpy.ones((n0, 1), dtype=dtype)
        n1 = v.nvec()
        e1 = numpy.ones((n1, 1), dtype=dtype)
        n = n0 + n1

        if shift:
            mean0 = self.__mean_v.data()
            mean1 = v.new_vectors(1, v.dimension())
            v.multiply(e1, mean1)
            mean1 = mean1.data()/n1
            mean = (n0/n)*mean0 + (n1/n)*mean1
            diff = mean0 - mean
            vdiff = v.new_vectors(diff)
            vdiff0 = vdiff.orthogonalize(right0)
            diff0 = vdiff0.data().T
            s = nla.norm(vdiff.data())*e0[:1]
            vdiff.scale(s)
            e0v = v.new_vectors(e0.T)
            left0.add(e0v, 1.0, diff0)
            e0v.scale(s, multiply=True)
            left0.append(e0v)
            right0.append(vdiff)
            vmean = v.new_vectors(mean)
            v.add(vmean, -1.0, e1.T)
        else:
            mean = None
            vmean = None

        s = abs(v.dots(v))
        fnorm = math.sqrt(numpy.sum(s))
        maxl2norm = numpy.amax(numpy.sqrt(s))

        left1 = v.orthogonalize(right0)

        lra = LowerRankApproximation()
        if rank < 0:
            if norm == 'f':
                update_tol = -tol*fnorm
            elif norm == 'm':
                update_tol = -tol*maxl2norm
            else:
                update_tol = -tol*sigma0
            urank = max_rank*n1//(n0 + n1)
            lra.compute(data_matrix, opt, tol=update_tol, norm=norm, \
                max_rank=urank, verb=verb)
        else:
            urank = rank*n1//(n0 + n1)
            if verb > 0:
                print('computing new %d components...' % urank)
            lra.compute(data_matrix, opt, rank=urank, verb=verb)

        left11 = lra.left_v()
        right10 = lra.right_v()
        
        new = left11.nvec()
        left01 = left0.new_vectors(new)
        left01.zero()
        left0.append(left01)
        left1.append(left11)
        left0_data = left0.data()
        left1_data = left1.data()
        left0_data = numpy.concatenate((left0_data, left1_data), axis=1)
        left0 = left0.new_vectors(left0_data)
#        left0.append(left1, axis=1)
        right0.append(right10)
        self.__left_v = left0
        self.__right_v = right0

        wl = left0.new_vectors(left0.nvec())
        wr = right0.new_vectors(right0.nvec())
        H = right0.dot(right0)
        mu, x = sla.eigh(H)
        q = mu[0] #/mu[-1]
        if q < 0.5:
#            k = numpy.sum(mu < 0.5)
#            print(mu[: k + 1])
            _lra_ortho(left0, right0, wl, wr)
        else:
            G = left0.dot(left0)
            lmd, x = sla.eigh(-G, H)
            y = nla.inv(x.T)
            left0.multiply(y, wl)
            wl.copy(left0)
            right0.multiply(x, wr)
            wr.copy(right0)

        if rank < 0:
            ncomp = right0.nvec()
            e = numpy.ones((n, 1), dtype=dtype)
            if norm == 'f':
                r = left0.dots(left0)
                s = math.sqrt(numpy.sum(r))
            elif norm == 'm':
                r = left0.dots(left0, transp=True)
                s = numpy.amax(numpy.sqrt(abs(r)))
            else:
                s = sigma[0]
            if shift and False:
                a = vmean.dot(vmean)
                b = vmean.dot(right0)
                p = left0.new_vectors(1)
                left0.multiply(b, p)
                if norm == 'f':
                    q = left0.new_vectors(e.T)
                    c = q.dot(p)
                    s = math.sqrt(s*s + 2*c + n*a)
                elif norm == 'm':
                    p = p.data()
                    s = math.sqrt(numpy.amax(r + 2*p.T) + a)
            eps = s*tol/4
            if norm == 'm':
                errs = numpy.zeros((1, n))
            s = 0
            i = 1
            while i < ncomp:
                if norm == 'f':
                    s = math.sqrt(s*s + r[ncomp - i])
                elif norm == 'm':
                    left0.select(1, ncomp - i)
                    lft = left0.data()
                    errs += lft * lft
                    s = numpy.amax(numpy.sqrt(errs))
                else:
                    s = sigma[ncomp - i]
                if s > eps:
                    break
                i += 1
            i -= 1
            if i > 0:
                if verb > 0:
                    print('discarding %d components out of %d' % (i, ncomp))
                ncomp -= i
        else:
            ncomp = rank

        left0.select(ncomp)
        right0.select(ncomp)
        self.__left = None
        self.__right = None
        self.__mean = None
        if shift:
            self.__mean_v = vmean
        self.__rank = self.__left_v.nvec()
        self.__tol = tol
        self.__svtol = svtol
        self.__norm = norm
        self.__arch = data_matrix.arch()
        self.__dtype = data_matrix.data_type()
        if max_rank > 0 and self.__left_v.nvec() > max_rank:
            self.__left_v.select(max_rank)
            self.__right_v.select(max_rank)
        self.iterations += lra.iterations

    def icompute(self, data_matrix, batch_size, opt=Options(), rank=-1, \
                 tol=0, norm='f', max_rank=-1, svtol=1e-3, shift=False, \
                 arch='cpu', verb=0):
        '''
        Computes Lower Rank Approximation of data_matix incrementally:
            - computes LRA for the block data_matrix[:batch_size, :]
            - in a loop updates LRA using subsequent blocks of data_matrix
              batch_size rows high (or less, for the last block)

        Parameters
        ----------
        data_matrix : 2D numpy array
            Data matrix A.
        batch_size : int
            The batch size.

        Remaining parameters have the same meaning as for compute().
        '''
        data_size = data_matrix.shape[0]
        batch_size = min(batch_size, data_size)
        batch = 0
        if self.__rank == 0:
            if verb > 0:
                print('processing batch %d of size %d' % (batch, batch_size))
            matrix = AMatrix(data_matrix[:batch_size, :], arch=arch)
            self.compute(matrix, opt=opt, rank=rank, \
                         tol=tol, norm=norm, max_rank=max_rank, svtol=svtol, \
                         shift=shift, verb=verb)
            first = batch_size
            batch += 1
        else:
            first = 0
        while first < data_size:
            next_ = min(data_size, first + batch_size)
            if verb > 0:
                print('processing batch %d of size %d' % (batch, next_ - first))
            matrix = AMatrix(data_matrix[first : next_, :], arch=arch, \
                             copy_data=True)
            self.update(matrix, opt=opt, rank=rank, tol=tol, norm=norm, \
                        max_rank=max_rank, svtol=svtol, verb=verb)
            first = next_
            batch += 1

    def mean(self): # mean row of A
        if self.__mean is None:
            if self.__mean_v is not None:
                self.__mean = self.__mean_v.data()
        return self.__mean

    def left(self): # left multiplier L
        if self.__left is None:
            if self.__left_v is not None:
                self.__left = self.__left_v.data().T
        return self.__left

    def right(self): # right multiplier R
        if self.__right is None:
            if self.__right_v is not None:
                self.__right = self.__right_v.data()
        return self.__right

    def mean_v(self):
        return self.__mean_v

    def left_v(self):
        return self.__left_v

    def right_v(self):
        return self.__right_v


class _DefaultLRAConvergenceCriteria:

    def __init__(self, tol):
        self.tolerance = tol
    def set_tolerance(self, tolerance):
        self.tolerance = tolerance
    def satisfied(self, solver, i):
        res = solver.convergence_data('residual', i)
        lmd = solver.convergence_data('eigenvalue', i)
        lmd_max = solver.convergence_data('max eigenvalue', i)
        tol = abs(lmd/lmd_max)**1.5*self.tolerance
        return res >= 0 and res*res <= tol


def _conj(a):
    if a.dtype.kind == 'c':
        return a.conj()
    else:
        return a


def _lra_ortho(v, u, wv, wu):

    u.copy(wu)
    s, q = wu.svd()
    v.multiply(_conj(q.T), wv)
    wv.scale(s, multiply=True)
    wv.copy(v)
    s, q = v.svd()
    wu.multiply(_conj(q.T), u)
    v.scale(s, multiply=True)



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

'''Compares images with their PCA approximations computed by the script
   compute_eigenimages.py.

Usage:
    show_errors [--help | -h] <images> <eigenimages>

Arguments:
    images       images file (must end with .npy)
    eigenimages  eigenimages file (must end with .npz)
'''

try:
    from docopt import docopt
    __version__ = '0.1.0'
    have_docopt = True
except:
    have_docopt = False

import math
import numpy
import numpy.linalg as nla
import pylab
import sys

try:
    input = raw_input
except NameError:
    pass


def _norm(a, axis):
    return numpy.apply_along_axis(nla.norm, axis, a)


if have_docopt:
    __version__ = '0.1.0'
    args = docopt(__doc__, version=__version__)
    img_file = args['<images>']
    eigim_file = args['<eigenimages>']
else:
    narg = len(sys.argv)
    if narg < 3:
        print('Usage: show_errors <images_file> <eigenimages_file>')
    img_file = sys.argv[1]
    eigim_file = sys.argv[2]

eigim = numpy.load(eigim_file)
u = eigim['eigim']
v = eigim['coord']
mean = eigim['mean']
nc, nyu, nxu = u.shape
m = v.shape[0]
print('%d eigenimages of size %dx%d loaded' % (nc, nyu, nxu))

print('loading images from %s...' % img_file)
images = numpy.load(img_file)
ni, ny, nx = images.shape
print('%d images of size %dx%d loaded' % (ni, ny, nx))
vmax = numpy.amax(images)

if m > ni:
    m = ni
if nx != nxu or ny != nyu:
    raise ValueError('mismatching image sizes (%d, %d) and (%d, %d)' % \
          (ny, nx, nyu, nxu))

sigma = _norm(v, axis=0)

pylab.figure()
pylab.plot(numpy.arange(1, nc + 1, 1), sigma)
pylab.xscale('log')
pylab.yscale('log')
pylab.grid()
pylab.title('singular values')
pylab.show()

n = nx*ny
images = numpy.reshape(images, (ni, n))
u = numpy.reshape(u, (nc, n))

print('measuring PCA errors...')
norm = _norm(images[:m, :], axis=1)
ones = numpy.ones((m, 1))
mean = numpy.reshape(mean, (1, n))
pca_images = numpy.dot(v, u) + numpy.dot(ones, mean)
mean = numpy.reshape(mean, n)
norm = vmax*math.sqrt(n)
err = _norm(images[:m, :] - pca_images, axis=1)/norm
ind = numpy.argsort(-err)
pylab.figure()
pylab.plot(numpy.arange(1, m + 1, 1), err[ind])
pylab.title('PCA errors')
pylab.show()
k = 100
print('%d images with the largest PCA error:' % k)
print(ind[:k])

while True:
    i = int(input('image number (negative to exit): '))
    if i < 0 or i >= ni:
        break
    image = numpy.reshape(images[i,:], (ny, nx))
    print('PCA error: %.1e' % err[i])
    if i < m:
        w = v[i, :]
    else:
        w = numpy.dot(u, images[i, :])
    pylab.figure()
    pylab.title('image %d' % i)
    img = images[i, :]
    image = numpy.reshape(img, (ny, nx))
    pylab.imshow(image, cmap='gray')
    img = numpy.dot(w, u)
    if mean is not None:
        img += mean
    pca_image = numpy.reshape(img, (ny, nx))
    pylab.figure()
    pylab.title('PCA approximation of the image')
    pylab.imshow(pca_image, cmap='gray')
    pylab.show()

print('done')



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="raleigh",
    version="1.1.4",
    author="Evgueni Ovtchinnikov",
    author_email="evgueni.ovtchinnikov@stfc.ac.uk",
    description="RAL eigensolver for real symmetric and Hermitian problems",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/evgueni-ovtchinnikov/raleigh",
    packages=setuptools.find_packages(),
    license = "BSD-3-Clause",
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

# -*- coding: utf-8 -*-
"""
Principal Components update demo.

Performs PCA on a chunk of data, then addds more data and updates principal 
components.

Usage: pca_update <data_file> <tolerance> <q_first> <max_pcs> [gpu]

data_file : the name of the file containing data matrix X
tolerance : PCA approximation tolerance wanted
q_first   : relative size of the first chunk
max_pcs   : maximal number of principal components to compute (<1: no limit)
gpu       : run on GPU if this argument is present
"""

import numpy
import sys
import timeit

from raleigh.interfaces.pca import pca, pca_error


narg = len(sys.argv)
if narg < 5:
    usage = \
          'Usage: pca_update <data_file> <tolerance> <q_first> <max_pcs> [gpu]'
    raise SystemExit(usage)
data = numpy.load(sys.argv[1])
atol = float(sys.argv[2])
q = float(sys.argv[3])
mpc = int(sys.argv[4])
arch = 'cpu' if narg < 6 else 'gpu!'

numpy.random.seed(1) # make results reproducible

m_all = data.shape[0]
n = data.shape[1]
if len(data.shape) > 2: # allow for multi-dimensional samples (e.g. images)
    n = numpy.prod(data.shape[1:])
    data = numpy.reshape(data, (m_all, n))
m = min(m_all, max(1, int(q*m_all)))

print('computing PCs for %d data samples...' % m)
start = timeit.default_timer()
mean, trans, comps = pca(data[: m, :], tol=atol, mpc=mpc, arch=arch)
elapsed = timeit.default_timer() - start
ncomp = comps.shape[0]
print('%d principal components computed in %.2e sec' % (ncomp, elapsed))
em, ef = pca_error(data[: m, :], mean, trans, comps)
print('PCA error: max %.1e, Frobenius %.1e' % (em, ef))

if m < m_all:
    print('\nmore data arrived, updating PCs for %d data samples...' % m_all)
    start = timeit.default_timer()
    mean, trans, comps = pca(data[m :, :], mpc=mpc, tol=atol, verb=0, arch=arch, \
        have=(mean, trans, comps))
    elapsed = timeit.default_timer() - start
    ncomp = comps.shape[0]
    print('%d principal components updated in %.2e sec' % (ncomp, elapsed))
    em, ef = pca_error(data, mean, trans, comps)
    print('PCA error: max %.1e, Frobenius %.1e' % (em, ef))

print('done')



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

"""Computes several eigenvalues and eigenvectors of a real symmetric matrix.

--------------------------------------------------------------------------------
Requires MKL 10.3 or later (needs mkl_rt.dll on Windows, libmkl_rt.so on Linux).
--------------------------------------------------------------------------------

Visit https://sparse.tamu.edu/ to download matrices (in Matrix Market format)
to test on. Recommended group: DNVS. Best performance compared to scipy eigsh
is achieved on matrices with large clusters such as DNVS/shipsec1 and in
low-to-moderate accuracy in eigenvectors (1e-3 to 1e-6, corresponds to single
to double machine precision in eigenvalues).

Usage:
    sparse_evp [--help | -h | options]

Options:
    -a <mat>, --matrix=<mat>  problem matrix (name of a Matrix Market file or
                              lap3d) [default: lap3d]
    -b <mss>, --mass=<mss>    mass matrix if problem is generalized
    -n <dim>, --dim=<dim>     mesh sizes nx = ny = nz for lap3d [default: 10]
    -s <sft>, --shift=<sft>   shift (ignored in preconditioned mode) [default: 0]
    -l <lft>, --left=<lft>    number of eigenvalues left of shift (shift-invert
                              mode) [default: 0]
    -r <rgt>, --right=<rgt>   number of eigenvalues right of shift (shift-ivert
                              mode) [default: 6]
    -e <nev>, --nev=<nev>     number of eigenvalues wanted: nearest to the shift
                              in shift-invert mode if left and right are zeros
                              or smallest if preconditioning is used
                              [default: 6]
    -t <tol>, --tol=<tol>     error/residual tolerance [default: 1e-6]
    -v <vrb>, --verb=<vrb>    verbosity level (negative suppresses all output)
                              [default: 0]
    -I, --invop  first argument of partial_hevp is a SparseSymmetricSolver
    -P, --ilutp  use mkl dcsrilut (incomplete ILU) as a preconditioner
    -C, --check  check the eigenvector errors by solving twice
"""

try:
    from docopt import docopt
    __version__ = '0.1.0'
    have_docopt = True
except:
    have_docopt = False

import numpy
from scipy.io import mmread
import scipy.sparse as scs
import sys
import time

from raleigh.algebra import verbosity
verbosity.level = 1

try:
    from raleigh.algebra.sparse_mkl import SparseSymmetricSolver
    from raleigh.algebra.sparse_mkl import IncompleteLU
    from raleigh.interfaces.partial_hevp import partial_hevp
except:
    print('This script requires MKL 10.3 or later, sorry!')
    exit()


def lap1d(n, a):
    h = a/(n + 1)
    d = numpy.ones((n,))/(h*h)
    return scs.spdiags([-d, 2*d, -d], [-1, 0, 1], n, n, format='csr')


def lap2d(nx, ny, ax, ay):
    L = lap1d(nx, ax)
    Ly = lap1d(ny, ay)
    L = scs.csr_matrix(scs.kron(scs.eye(ny), L) + scs.kron(Ly, scs.eye(nx)))
    return L


def lap3d(nx, ny, nz, ax, ay, az):
    L = lap2d(nx, ny, ax, ay)
    Lz = lap1d(nz, az)
    L = scs.csr_matrix(scs.kron(scs.eye(nz), L) + scs.kron(Lz, scs.eye(nx*ny)))
    return L


def norm(a, axis):
    return numpy.apply_along_axis(numpy.linalg.norm, axis, a)


def vec_err(u, v, B=None):
    if B is None:
        w = v.copy()
    else:
        w = B.dot(v)
    q = numpy.dot(u.T, w)
    w = numpy.dot(u, q) - v
    s = norm(w, axis = 0)
    return s


if have_docopt:
    args = docopt(__doc__, version=__version__)
    matrix = args['--matrix']
    mass = args['--mass']
    if matrix == 'lap3d':
        nx = int(args['--dim'])
    sigma = float(args['--shift'])
    left = int(args['--left'])
    right = int(args['--right'])
    nev = int(args['--nev'])
    tol = float(args['--tol'])
    verb = int(args['--verb'])
    invop = args['--invop']
    ilutp = args['--ilutp']
    check = args['--check']
else:
    print('\n=== docopt not found, using default options...\n')
    matrix = 'lap3d'
    nx = 10
    mass = None
    sigma = 0
    left = 0
    right = 6
    nev = 6
    tol = 1e-10
    verb = 0
    invop = False
    ilutp = False
    check = False

numpy.random.seed(1) # makes the results reproducible

if matrix == 'lap3d':
    if verb > -1:
        print('generating discretized 3D Laplacian matrix...')
    M = lap3d(nx, nx, nx, 1.0, 1.0, 1.0)
    ia = M.indptr + 1
    n = ia.shape[0] - 1
    if mass is not None: # just a simple generalized problem test
        B = 2*scs.eye(n, format='csr')
    else:
        B = None
else:
    if verb > -1:
        print('reading the matrix from %s...' % matrix)
    M = mmread(matrix).tocsr()
    n = M.shape[0]
    if mass is not None:
        if verb > -1:
            print('reading the mass matrix from %s...' % mass)
        B = mmread(mass).tocsr()
    else:
        B = None

T = None

if invop:
    if verb > -1:
        print('setting up the linear system solver...')
    start = time.time()
    A = SparseSymmetricSolver()
    A.analyse(M, sigma, B)
    A.factorize()
    stop = time.time()
    setup_time = stop - start
    if verb > -1:
        print('setup time: %.2e' % setup_time)
else:
    A = M
    if ilutp:
        if verb > -1:
            print('setting up the preconditioner...')
        start = time.time()
        T = IncompleteLU(M)
        T.factorize()
        stop = time.time()
        setup_time = stop - start
        if verb > -1:
            print('setup time: %.2e' % setup_time)

if T is not None or left == 0 and right == 0:
    which = nev
else:
    which = (left, right)

vals, vecs, status = partial_hevp(A, B, T, sigma=sigma, which=which, tol=tol, \
                                  verb=verb)
if status != 0 and verb > -1:
    print('partial_hevp execution status: %d' % status)
if verb > -1:
    print('converged eigenvalues are:')
    print(vals)

if check:
    '''Estimate eigenvector errors by solving twice.

    Rationale:
    Consider for simplicity the case of one eigenvector computed.
    Let u1 and u2 be the two approximations to a given eigenvector u computed
    by two calls to partial_evp. Since the initial eigenvector guesses are
    random, u1 and u2 are different, and errors (I - P)u1 and (I - P)u2,
    where P is the orthogonal projector onto u and I is the identity, are
    essentially random vectors, and hence nearly orthogonal. Hence, the norm
    of u2 - u1 cannot be significantly less than the largest of the two errors
    and, at the same time, cannot be significantly greater than this error
    multiplied by the square root of 2, i.e. is very close to the actual error.
    '''
    nev = vals.shape[0]
    lft = numpy.sum(vals < sigma)
    rgt = nev - lft
    vls, vcs, status = partial_hevp(A, B, T, sigma=sigma, which=which, tol=tol, \
                                    verb=verb)
    if status != 0 and verb > -1:
        print('partial_hevp execution status: %d' % status)
    ne = vls.shape[0]
    lt = numpy.sum(vls < sigma)
    rt = ne - lt
    left = min(lft, lt)
    right = min(rgt, rt)
    print('eigenvector errors:')
    if left > 0:
        errs = vec_err(vecs[:, lft - left : lft], vcs[:, lt - left : lt], B)
        print(errs)
    if right > 0:
        errs = vec_err(vecs[:, lft : lft + right], vcs[:, lt : lt + right], B)
        print(errs)



# Copyright 2019 United Kingdom Research and Innovation 
# Author: Evgueni Ovtchinnikov (evgueni.ovtchinnikov@stfc.ac.uk)

# -*- coding: utf-8 -*-
""" 
'Smart' Principal Component Analysis test.

Computes PCA approximation of specified accuracy.

PCA allows to approximate m data samples of size n represented by rows of an 
m-by-n matrix X by the rows of the matrix X_k = e a + Y_k Z_k, where 
    e = numpy.ones((m, 1)),
    a = numpy.mean(X, axis=0).reshape((1, n)), 
    the rows of m-by-k matrix Y_k are reduced-features data samples, and 
    the rows of k-by-n matrix Z_k are principal components.

This script computes PCA approximation X_k such that the Frobenius norm of the
difference X - X_k is not greater than the user-specified tolerance.

If sklearn is installed, compares with restarted sklearn.decomposition.PCA.

Usage: pca_smart <data_file> <tolerance> <max_pcs> [gpu]

data_file : the name of the file containing data matrix X
tolerance : approximation tolerance wanted
max_pcs   : maximal number of principal components to compute (<1: no limit)
gpu       : run raleigh pca on GPU if this argument is present
"""

import math
import numpy
import numpy.linalg as nla
import sys
import timeit

from raleigh.interfaces.pca import pca, pca_error


narg = len(sys.argv)
if narg < 4:
    usage = 'Usage: pca_smart <data_file> <tolerance> <max_pcs> [gpu]'
    raise SystemExit(usage)
data = numpy.load(sys.argv[1])
m = data.shape[0]
n = numpy.prod(data.shape[1:])
if len(data.shape) > 2:
    data = numpy.reshape(data, (m, n))
tol = float(sys.argv[2])
mpc = int(sys.argv[3])
arch = 'cpu' if narg < 5 else 'gpu!'

numpy.random.seed(1) # make results reproducible

print('\n--- solving with raleigh pca...\n')
start = timeit.default_timer()
mean, trans, comps = pca(data, tol=tol, mpc=mpc, arch=arch, verb=0)
elapsed = timeit.default_timer() - start
ncomp = comps.shape[0]
print('%d principal components computed in %.2e sec' % (ncomp, elapsed))
em, ef = pca_error(data, mean, trans, comps)
print('PCA error: max 2-norm %.1e, Frobenius norm %.1e' % (em, ef))

try:
    from sklearn.decomposition import PCA
    '''
    sklearn PCA needs to be given the number k of wanted components;
    since it is generally unknown in advance, one has to apply deflation:
    a certain number k of principal components and corresponding 
    reduced-features samples are computed, and if the PCA error E_k = X - X_k
    is not small enough, then further k PCs of X (which are PCs of E_k) are 
    computed etc. until the PCA error falls below the required tolerance
    '''
    print('\n--- solving with sklearn PCA...')
    dtype = data.dtype
    # a guess of sufficient number of PCs
    if mpc > 0:
        npc = mpc//2
    else:
        npc = max(1, min(m, n)//10)
    data0 = data # data will change, save a copy for final error computation
    data = data.copy()
    start = timeit.default_timer()
    skl_pca = PCA(npc)
    sigma_skl = numpy.ndarray((0,), dtype=dtype)
    comps_skl = numpy.ndarray((0, n), dtype=dtype)
    trans_skl = numpy.ndarray((m, 0), dtype=dtype)
    ones = numpy.ones((m, 1), dtype=data.dtype)
    mean = numpy.dot(ones.T, data)/m
    data_s = data - numpy.dot(ones, mean)
    # compute Frobenius norm of the initial PCA error
    nrms = nla.norm(data_s, axis=1)
    err2 = numpy.sum(nrms*nrms)
    err2init = err2
    while True:
        # compute next portion of PCs
        trans = skl_pca.fit_transform(data)
        sigma = skl_pca.singular_values_
        comps = skl_pca.components_
        sigma_skl = numpy.concatenate((sigma_skl, sigma))
        comps_skl = numpy.concatenate((comps_skl, comps))
        trans_skl = numpy.concatenate((trans_skl, trans), axis=1)
        stop = timeit.default_timer()
        time_s = stop - start
        pcs = comps_skl.shape[0]
        # update Frobenius norm of PCA error
        err2 -= numpy.sum(sigma*sigma)
        err = math.sqrt(err2/err2init)
        print('%.2f sec: last singular value: sigma[%d] = %e, error %.2e' \
            % (time_s, pcs - 1, sigma[-1], err))
        if err < tol: # desired accuracy achieved, quit the loop
            break
        if mpc > 0 and pcs >= mpc: # max number of PCs exceeded
            break
        print('deflating...')
        # deflate: subtract computed approximation from data
        data -= numpy.dot(trans, comps)
        print('restarting...')
    elapsed = timeit.default_timer() - start
    ncomp = comps_skl.shape[0]
    print('%d principal components computed in %.2e sec' % (ncomp, elapsed))
    em, ef = pca_error(data0, mean, trans_skl, comps_skl)
    print('PCA error: max 2-norm %.1e, Frobenius norm %.1e' % (em, ef))
except:
    pass

print('done')
